import re


def update_credentials(credentials, token, cookies):
    credentials.token = token
    credentials.cookies = cookies
    credentials.invalid_creds = False
    return credentials


def get_formatted_bot_name(name):
    bot_name = name
    bot_name.replace(" ", "_") \
        .replace(".", "dot_") \
        .replace("#", "") \
        .replace("+", "_plus")

    return re.sub(r'[^A-Za-z0-9]', "", bot_name).lower()



import json
from datetime import datetime, UTC
from lgt_jobs.lgt_common.pubsub.messages import publish_message2_pubsub
from lgt_jobs.lgt_data.engine import DelayedJob
from .basejobs import InvalidJobTypeException, BaseBackgroundJobData, BaseBackgroundJob
from .env import background_jobs_topic


def datetime_converter(o):
    if isinstance(o, datetime):
        return o.__str__()


class BackgroundJobRunner:
    @staticmethod
    def run(jobs_map: dict, data: dict):
        """
        @:param data received after dump
        @:param jobs_map job instance mapping
        """
        job_type_name = data["job_type"]
        job = jobs_map.get(job_type_name, None)

        if not job:
            raise InvalidJobTypeException(f"Unable to find job '{job_type_name}' in the list of modules")

        return job().run(data["data"])

    @staticmethod
    def submit(job: type, data: BaseBackgroundJobData):
        job_data = BaseBackgroundJob.dumps(job, data.dict())
        BackgroundJobRunner.submit_data(job_data)

    @staticmethod
    def submit_data(data: dict):
        json_str = json.dumps(data, ensure_ascii=False, default=datetime_converter)
        BackgroundJobRunner.submit_json(json_str)

    @staticmethod
    def submit_json(json_str: str):
        publish_message2_pubsub(background_jobs_topic, message_json=json_str)

    @staticmethod
    def schedule(jib: str, job: type, data: BaseBackgroundJobData, scheduled_at: datetime):
        job_data = BaseBackgroundJob.dumps(job, data.model_dump())
        DelayedJob(
            jib=jib,
            created_at=datetime.now(UTC),
            scheduled_at=scheduled_at,
            job_type=job.__name__,
            data=json.dumps(job_data, ensure_ascii=False, default=datetime_converter)
        ).save()

    @staticmethod
    def reschedule(jib: str, job: type, data: BaseBackgroundJobData, scheduled_at: datetime):
        BackgroundJobRunner.remove_schedule(jib)
        BackgroundJobRunner.schedule(jib, job, data, scheduled_at)

    @staticmethod
    def remove_schedule(jib: str):
        [job.delete() for job in DelayedJob.objects(jib=jib)]



import loguru
import requests
from requests import Response
from lgt_jobs.lgt_common.discord_client.methods import DiscordMethods


class DiscordClient:
    base_url = 'https://discord.com/api/'
    discord_api_version = 'v9/'
    token: str
    headers: dict

    def __init__(self, token: str = None):
        self.token = token
        self.headers = {"Authorization": self.token}

    def login(self, login: str, password: str, captcha_key: str = None) -> dict:
        payload = {
            'login': login,
            'password': password,
            'captcha_key': captcha_key,
            "undelete": False,
            "login_source": "",
            "gift_code_sku_id": None
        }
        response = requests.post(f"{self.base_url}{self.discord_api_version}{DiscordMethods.LOGIN.value}", json=payload)
        if response.status_code == 400 or response.status_code == 200:
            return response.json()
        return {}

    def get_servers(self) -> list | dict:
        response = requests.get(f"{self.base_url}{DiscordMethods.USER_GUILDS.value}", headers=self.headers)
        return self.__response(response).json()

    def get_dms(self) -> list | dict:
        response = requests.get(f"{self.base_url}{DiscordMethods.USER_DMS.value}", headers=self.headers)
        return self.__response(response).json()

    def get_current_user(self) -> dict:
        response = requests.get(f"{self.base_url}{DiscordMethods.USER.value}", headers=self.headers)
        if response.status_code != 200:
            self.__log_error(response)
        return response.json()

    def get_channels(self, guild_id: str) -> list | dict:
        response = requests.get(f"{self.base_url}{DiscordMethods.guild_channels(guild_id)}", headers=self.headers)
        return self.__response(response).json()

    def get_invite_link(self, channel_id: str) -> Response:
        response = requests.post(f'{self.base_url}{self.discord_api_version}'
                                 f'{DiscordMethods.channels_invites(channel_id)}', headers=self.headers)
        return self.__response(response)

    @staticmethod
    def __log_error(response: Response):
        loguru.logger.warning(f"[DiscordClient WARNING]: {response.url}, {response.status_code}, {response.content}")

    @staticmethod
    def __response(response: Response):
        if response.status_code != 200:
            DiscordClient.__log_error(response)
        return response



import sys
from loguru import logger as log

# log.remove()
# log.add(sys.stdout, format="{time} {name} {level} {message}", level="INFO",
#         filter=lambda record: record['level'].name == 'INFO')
# log.add(sys.stderr, format="{time} {name} {level} {message}", level="WARNING",
#         filter=lambda record: record['level'].name == 'WARNING')
# log.add(sys.stderr, format="{time} {name} {level} {message}", level="ERROR",
#         filter=lambda record: record['level'].name in ['ERROR', 'FATAL', 'CRITICAL'])
#
#
# log.info('Logger has been configured')
#
# log = log



from __future__ import annotations
import copy
import json
from abc import ABC
from datetime import datetime, UTC
from typing import Optional, List, Dict
from .enums import UserRole, SourceType, FeaturesEnum, FeatureOptions
from bson import ObjectId


class DictionaryModel(ABC):
    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        model = cls()
        for k, v in dic.items():
            setattr(model, k, v)

        if '_id' in dic:
            setattr(model, 'id', dic['_id'])

        return model

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)
        return result


class BaseModel(DictionaryModel):
    def __init__(self):
        self.id = None
        self.created_at = datetime.now(UTC)


class Credentials(BaseModel):
    def __init__(self):
        super().__init__()
        self.token = None
        self.cookies = None
        self.invalid_creds = False


class Source:
    def __init__(self):
        self.source_type: SourceType | None = None
        self.source_name = None
        self.source_id = None

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        model = cls()
        for k, v in dic.items():
            setattr(model, k, v)

        return model

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)
        return result


class BaseConfig:
    def __init__(self):
        self.owner = None
        self.id = None

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        model = cls()
        for k, v in dic.items():
            setattr(model, k, v)

        return model

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)
        return result


class Config(BaseConfig):
    def __init__(self):
        super().__init__()
        self.name = None


class BaseBotModel(Credentials):
    def __init__(self):
        super().__init__()
        self.created_by = None
        self.user_name = None
        self.slack_url = None
        self.registration_link = None
        self.channels = None
        self.connected_channels = None
        self.channels_users = None
        self.users_count = None
        self.messages_received: int = 0
        self.messages_filtered: int = 0
        self.recent_messages: List[str] = []
        self.icon = None
        self.active_channels = {}
        self.paused_channels = []
        self.source: Source | None = None
        self.two_factor_required: bool = False
        self.banned: bool = False
        self.associated_user = None
        self.type: SourceType | None = None
        self.deleted = False


class DedicatedBotModel(BaseBotModel):
    def __init__(self):
        super().__init__()
        self.user_id: Optional[str] = None
        self.updated_at: Optional[datetime] = datetime.now(UTC)
        self.servers: List[Server] = []
        self.state = 0
        self.source: Source | None = None

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)

        if result.get('source'):
            result['source'] = Source.to_dic(result.get('source'))

        result['servers'] = [Server.to_dic(server) for server in self.servers]
        return result

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        model: DedicatedBotModel = cls()
        for k, v in dic.items():
            setattr(model, k, v)

        if '_id' in dic:
            setattr(model, 'id', dic['_id'])

        model.source = Source.from_dic(dic.get("source"))
        model.servers = [Server.from_dic(server) for server in dic.get("servers", [])]
        return model


class Server:
    pass

    def __init__(self):
        self.id = None
        self.name = None
        self.channels: List[Channel] = []
        self.icon = None
        self.active = False
        self.deleted = False
        self.subscribers = 0

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        model = cls()
        for k, v in dic.items():
            if hasattr(model, k):
                setattr(model, k, v)

        model.channels = [Channel.from_dic(channel) for channel in dic.get("channels", [])]
        return model

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)
        result['channels'] = [Channel.to_dic(channel) for channel in self.channels]
        return result


class Channel:
    pass

    def __init__(self):
        self.id = None
        self.name = None
        self.type = None
        self.is_member = True
        self.active = True
        self.subscribers = 0

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        model = cls()
        for k, v in dic.items():
            if hasattr(model, k):
                setattr(model, k, v)

        return model

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)
        return result


class MessageModel:
    pass

    def __init__(self):
        self.message_id = None
        self.channel_id = None
        self.channel_name = None
        self.message = None
        self.name = None
        self.sender_id = None
        self.source: Source | None = None
        self.companies: List[str] = list()
        self.technologies: List[str] = list()
        self.locations: List[str] = list()
        self.configs: List[BaseConfig] = list()
        self.attachments: List[dict] = []
        self.timestamp = None
        self.tags: List[str] = []

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None
        if isinstance(dic.get('attachments'), str):
            dic['attachments'] = json.loads(dic['attachments'])

        model: MessageModel = cls()
        for k, v in dic.items():
            setattr(model, k, v)

        model.source = Source.from_dic(dic.get("source"))
        model.configs = [BaseConfig.from_dic(doc) for doc in dic.get("configs", [])]
        return model

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)

        if result.get('source'):
            result['source'] = Source.to_dic(result.get('source'))
        if result.get('configs'):
            result['configs'] = [BaseConfig.to_dic(config) for config in result.get('configs')]
        return result


class UserModel(BaseModel):
    def __init__(self):
        super().__init__()
        self.email: Optional[str] = None
        self.password: Optional[str] = None
        self.roles: List[str] = []
        self.user_name: str = ''
        self.company: str = ''
        self.company_size: Optional[int] = None
        self.company_industries: Optional[List[str]] = None
        self.company_technologies: Optional[List[str]] = None
        self.company_locations: Optional[List[str]] = None
        self.company_web_site: str = ''
        self.company_description: str = ''
        self.position: str = ''
        self.new_message_notified_at: Optional[datetime] = None
        self.photo_url: str = ''
        self.slack_profile = SlackProfile()
        self.leads_limit: Optional[int] = None
        self.leads_proceeded: Optional[int] = None
        self.leads_filtered: Optional[int] = None
        self.leads_limit_updated_at: Optional[int] = None
        self.keywords: Optional[List[str]] = None
        self.block_words: Optional[List[str]] = None
        self.paid_lead_price: int = 1
        self.state: int = 0
        self.credits_exceeded_at: Optional[datetime] = None
        self.unanswered_leads_period = None
        self.inactive = None
        self.slack_users: List[SlackUser] = []
        self.discord_users: List[DiscordUser] = []
        self.verified: bool = False
        self.subscription_id: ObjectId | None = None
        self.subscription_expired_at: datetime | None = None
        self.balance: str | None = None
        self.subscription_name: str | None = None

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        model: UserModel = cls()
        for k, v in dic.items():
            setattr(model, k, v)

        if '_id' in dic:
            setattr(model, 'id', dic['_id'])

        model.slack_profile = SlackProfile.from_dic(dic.get('slack_profile'))
        model.slack_users = [SlackUser.from_dic(user) for user in dic.get('slack_users', [])]
        model.discord_users = [DiscordUser.from_dic(user) for user in dic.get('discord_users', [])]
        return model

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)

        if result.get('slack_profile', None):
            result['slack_profile'] = result.get('slack_profile').__dict__

        return result

    @property
    def is_admin(self):
        return UserRole.ADMIN in self.roles

    def get_slack_user(self, slack_email: str):
        return next(filter(lambda x: slack_email == x.email, self.slack_users), None)

    def get_discord_user(self, login: str):
        return next(filter(lambda x: login == x.login, self.discord_users), None)


class DiscordUser(BaseModel):
    pass

    def __init__(self):
        super().__init__()
        self.created_at = datetime.now(UTC)
        self.login = ''
        self.captcha_key = ''
        self.status = None
        self.workspaces: List[UserWorkspace] = []

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)

        if result.get('workspaces', None):
            result['workspaces'] = [ws.__dict__ for ws in result.get('workspaces')]

        return result

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        model = cls()
        for k, v in dic.items():
            setattr(model, k, v)

        model.workspaces = [UserWorkspace.from_dic(ws) for ws in dic.get('workspaces', [])]
        return model


class SlackUser:
    pass

    def __init__(self):
        self.created_at = datetime.now(UTC)
        self.cookies = {}
        self.email = ''
        self.status = None
        self.workspaces: List[UserWorkspace] = []

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)

        if result.get('workspaces', None):
            result['workspaces'] = [ws.__dict__ for ws in result.get('workspaces')]

        return result

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        model = cls()
        for k, v in dic.items():
            setattr(model, k, v)

        model.workspaces = [UserWorkspace.from_dic(ws) for ws in dic.get('workspaces', [])]
        return model


class UserWorkspace:
    pass

    def __init__(self):
        super().__init__()
        self.id = ''
        self.name = ''
        self.url = ''
        self.domain = ''
        self.active_users = ''
        self.profile_photos = []
        self.associated_user = ''
        self.magic_login_url = ''
        self.magic_login_code = ''
        self.user_email = ''
        self.user_type = ''
        self.variant = ''
        self.token = ''
        self.icon = ''
        self.two_factor_required = False

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        model: UserWorkspace = cls()
        for k, v in dic.items():
            setattr(model, k, v)

        model.icon = dic.get('icon_88', "")
        return model


class UserResetPasswordModel(BaseModel):
    pass

    def __init__(self):
        super().__init__()
        self.email = None


class LeadModel(BaseModel):
    def __init__(self):
        super().__init__()
        self.status = ''
        self.notes = ''
        self.archived = False
        self.message: Optional[MessageModel] = None
        self.hidden = False
        self.followup_date = None
        self.score = 0
        self.board_id = None
        self.linkedin_urls = []
        self.likes = 0
        self.reactions = 0
        self.replies = []
        self.last_action_at: Optional[datetime] = None
        self.slack_channel = None

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        model: LeadModel = cls()
        for k, v in dic.items():
            setattr(model, k, v)

        model.message = MessageModel.from_dic(dic['message'])
        if not model.last_action_at:
            model.last_action_at = model.created_at

        return model

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)
        result["message"] = self.message.to_dic()
        result['archived'] = self.archived
        return result


class ExtendedLeadModel(LeadModel):
    def __init__(self):
        super().__init__()
        self.previous_publications = []
        self.last_conversation: List[ChatMessage] = []
        self.contact: SlackMemberInformation | None = None
        self.deleted = False
        self.user_lead: UserLeadModel | None = None
        self.dedicated: bool = False
        self.bots: List[BotInfo] = []
        self.user_contact: UserContact | None = None
        self.paid: bool = False

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        result: ExtendedLeadModel | None = LeadModel.from_dic(dic)
        if not result:
            return None

        result.contact = SlackMemberInformation.from_dic(dic.get('contact'))
        result.user_contact = UserContact.from_dic(dic.get('user_contact'))
        result.previous_publications = [LeadModel.from_dic(lead) for lead in dic.get('previous_publications', [])]
        result.user_lead = UserLeadModel.from_dic(dic.get('user_lead'))
        result.last_conversation = [ChatMessage.from_dic(message) for message in dic.get('last_conversation', [])]
        result.bots = [BotInfo.from_dic(bot) for bot in dic.get('bots', [])]
        return result

    def to_csv(self, board_name: str) -> List[str]:
        return [self.message.source, self.contact.real_name, self.contact.title, self.contact.email,
                self.notes, board_name, self.status,
                self.followup_date.strftime("%d.%m.%Y %H:%M") if self.followup_date else "",
                self.message.message.replace('\n', ' ').strip()]


class BotInfo:
    def __init__(self):
        self.id = None
        self.invalid_creds: bool | None = False
        self.source = None
        self.banned: bool | None = False
        self.user_name: str = ''
        self.associated_user: str | None = ''
        self.deleted: bool = False
        self.two_factor_required = False

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        model = BotInfo()
        for k, v in dic.items():
            if hasattr(model, k):
                setattr(model, k, v)

        if '_id' in dic:
            setattr(model, 'id', dic['_id'])

        return model

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)
        return result


class SlackReplyModel(BaseModel):
    def __init__(self):
        super().__init__()
        self.type = None
        self.user = None
        self.username = None
        self.text = None
        self.thread_ts = None
        self.parent_user_id = None
        self.ts = None
        self.files = []
        self.attachments = []

    @classmethod
    def from_slack_response(cls, dic: dict):
        if not dic:
            return None

        model = cls()
        for k, v in dic.items():
            setattr(model, k, v)

        js_ticks = int(model.ts.split('.')[0] + model.ts.split('.')[1][3:])
        model.created_at = datetime.fromtimestamp(js_ticks / 1000.0)

        if model.files:
            model.files = [{"url_private_download": file.get("url_private_download")} for file in model.files]

        return model


class ChatMessage:
    bot_id: ObjectId
    user_id: ObjectId
    sender_id: str
    text: str
    user: str
    id: str
    viewed: bool
    files: list
    attachments: list[dict] | None
    created_at: datetime | None

    class SlackFileModel:
        def __init__(self):
            self.id = None
            self.name = None
            self.title = None
            self.filetype = None
            self.size = 0
            self.mimetype = None
            self.download_url = None

        def to_dic(self):
            result = copy.deepcopy(self.__dict__)
            return result

    def __init__(self):
        self.viewed = False
        self.text: str = ''
        self.created_at: datetime
        self.user = ''
        self.id = ''
        self.files = []
        self.attachments = []

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)
        if self.files and 'files' in result:
            result['files'] = [x.to_dic() if isinstance(x, ChatMessage.SlackFileModel) else x for x in self.files]

        return result

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None
        model = cls()
        for k, v in dic.items():
            setattr(model, k, v)
        return model


class ScheduledChatMessage(ChatMessage):
    post_at: Optional[datetime]
    jib: Optional[str]

    def __init__(self):
        super(ScheduledChatMessage, self).__init__()

        self.post_at = None
        self.jib = None


class UserLeadModel(LeadModel):
    pass

    def __init__(self):
        super().__init__()
        self.order: int = 0
        self.followup_date = None
        self.user_id = None
        self.chat_viewed_at = None
        self.board_id = None

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        result: UserLeadModel | None = super().from_dic(dic)
        if not result:
            return None

        result.chat_viewed_at = dic.get('chat_viewed_at')
        return result

    @staticmethod
    def from_route(lead: LeadModel):
        model_dict = lead.to_dic()
        result = UserLeadModel.from_dic(model_dict)
        result.order = 0

        result.message = MessageModel.from_dic(model_dict['message'])
        result.chat_viewed_at = None
        return result


class ExtendedUserLeadModel(UserLeadModel):
    pass

    def __init__(self):
        super().__init__()
        self.contact: SlackMemberInformation | None = None
        self.previous_publications = []
        self.bots: List[BotInfo] = []

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        result: ExtendedUserLeadModel | None = super().from_dic(dic)
        if not result:
            return None

        result.contact = SlackMemberInformation.from_dic(dic.get('contact'))
        result.previous_publications = [LeadModel.from_dic(lead) for lead in dic.get('previous_publications', [])]
        return result

    def to_dic(self):
        result = super().to_dic()
        result["contact"] = self.contact.to_dic()
        return result

    def to_csv(self, board_name: str) -> List[str]:
        return [self.message.source, self.contact.real_name, self.contact.title, self.contact.email,
                self.notes, board_name, self.status,
                self.followup_date.strftime("%d.%m.%Y %H:%M") if self.followup_date else "",
                self.message.message.replace('\n', ' ').strip()]


class BoardModel(BaseModel):
    pass

    def __init__(self):
        super().__init__()
        self.name = None
        self.user_id = None
        self.statuses: List[BoardedStatus] = []
        self.is_primary = None
        self.default = False

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        model = BoardModel()
        for k, v in dic.items():
            setattr(model, k, v)

        model.id = dic.get('_id')
        model.statuses = [BoardedStatus.from_dic(status) for status in dic.get('statuses', [])]
        return model

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)
        result["statuses"] = [BoardedStatus.to_dic(status) for status in self.statuses]

        for status in result['statuses']:
            status['board_id'] = result['id']

        return result


class BoardedStatus:
    pass

    def __init__(self):
        self.id = None
        self.name = None
        self.order = 0
        self.is_primary = False
        self.default = False
        self.user_leads = 0
        self.collapsed = False

    def to_dic(self):
        self.id = self.name
        return copy.deepcopy(self.__dict__)

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None
        model = cls()
        for k, v in dic.items():
            setattr(model, k, v)
        return model


class SlackProfile:
    pass

    def __init__(self):
        self.title = ''
        self.phone = ''
        self.skype = ''
        self.display_name = ''
        self.real_name = ''
        self.email = ''

    def to_dic(self):
        return copy.deepcopy(self.__dict__)

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None
        model = cls()
        for k, v in dic.items():
            setattr(model, k, v)
        return model


class SlackMemberInformation(BaseModel, SlackProfile):
    workspace: str
    sender_id: str
    images: dict
    full_text: str
    deleted: bool = False
    is_bot: bool = False
    is_app_user: bool = False
    is_admin: bool = False
    is_owner: bool = False
    is_email_confirmed: bool = False
    online: Optional[str] = None
    online_updated_at: datetime = None
    timezone: SlackTimeZone = None
    source: Source = None

    @classmethod
    def from_dic(cls, dic: dict):
        model: SlackMemberInformation = cls()
        if not dic:
            return None

        for k, v in dic.items():
            setattr(model, k, v)

        model.online = dic.get('online', '') == "active"
        model: SlackMemberInformation | None = super().from_dic(dic)
        model.source = Source.from_dic(dic.get('source'))
        return model

    def to_dic(self):
        result = copy.deepcopy(self.__dict__)
        if result.get('source'):
            result['source'] = Source.to_dic(result.get('source'))
        return result

    @staticmethod
    def from_slack_response(slack_profile: dict, source: Source = None):
        member_info: SlackMemberInformation = SlackMemberInformation()
        member_info.source = source
        member_info.sender_id = slack_profile.get("id")
        member_info.display_name = slack_profile["profile"].get("display_name")
        member_info.real_name = slack_profile["profile"].get("real_name")
        member_info.title = slack_profile["profile"].get("title")
        member_info.phone = slack_profile["profile"].get("phone")
        member_info.skype = slack_profile["profile"].get("skype")
        member_info.email = slack_profile["profile"].get("email")
        member_info.images = {
            'image_24': slack_profile.get("profile", {}).get("image_24",
                                                             'https://a.slack-edge.com/80588/img/slackbot_24.png'),
            'image_32': slack_profile.get("profile", {}).get("image_32",
                                                             'https://a.slack-edge.com/80588/img/slackbot_32.png'),
            'image_48': slack_profile.get("profile", {}).get("image_48",
                                                             'https://a.slack-edge.com/80588/img/slackbot_48.png'),
            'image_72': slack_profile.get("profile", {}).get("image_72",
                                                             'https://a.slack-edge.com/80588/img/slackbot_72.png'),
            'image_192': slack_profile.get("profile", {}).get("image_192",
                                                              'https://a.slack-edge.com/80588/img/slackbot_192.png'),
            'image_512': slack_profile.get("profile", {}).get("image_512",
                                                              'https://a.slack-edge.com/80588/img/slackbot_512.png'),

        }
        member_info.timezone = {"tz": slack_profile.get("tz"), "tz_label": slack_profile.get("tz_label"),
                                "tz_offset": slack_profile.get("tz_offset")}
        return member_info


class UserContact(SlackMemberInformation):
    chat_id: str
    user_id: ObjectId
    source_id: str
    scheduled_messages: List[ScheduledChatMessage] = []
    last_message_at: Optional[datetime]

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None
        model: UserContact | None = super().from_dic(dic)
        model.chat_history = [ChatMessage.from_dic(message) for message in dic.get('chat_history', [])]
        model.scheduled_messages = [ScheduledChatMessage.from_dic(item) for item in dic.get("scheduled_messages", [])]
        return model


class SlackTimeZone:
    tz: Optional[str]
    tz_label: Optional[str]
    tz_offset: Optional[int]


class ExtendedSlackMemberInformation(SlackMemberInformation):
    previous_publications = []
    name: str = None
    bots: List[BotInfo] = []

    @classmethod
    def from_dic(cls, dic: dict):
        model: ExtendedSlackMemberInformation | None = super().from_dic(dic)
        if not model:
            return None

        model.previous_publications = [LeadModel.from_dic(lead) for lead in dic.get('previous_publications', [])]
        model.bots = [BotInfo.from_dic(bot) for bot in dic.get('bots', [])]
        return model

    @staticmethod
    def to_lead(contact: ExtendedSlackMemberInformation, linkedin_contacts: Dict[str, LinkedinContact] = None) \
            -> ExtendedUserLeadModel:
        lead = ExtendedUserLeadModel()
        lead.id = str(contact.id)
        lead.created_at = contact.created_at
        lead.notes = ""
        lead.slack_channel = None
        lead.hidden = True
        lead.replies = []
        lead.reactions = 0
        lead.last_action_at = datetime.now(UTC)
        lead.created_at = datetime.now(UTC)
        if not hasattr(contact, "real_name"):
            contact.real_name = contact.name
        if not hasattr(contact, "display_name"):
            contact.display_name = contact.name

        lead.linkedin_urls = [f"https://www.linkedin.com/search/results/all/?keywords={contact.real_name}&origin=GLOBAL_SEARCH_HEADER&sid=u%40F"]
        lead.message = MessageModel()
        lead.message.message = contact.real_name
        if contact.title:
            lead.message.message += contact.title
        lead.message.message_id = str(contact.id)
        lead.message.name = contact.source.source_name
        lead.message.source = contact.source
        lead.message.sender_id = contact.sender_id
        lead.message.companies = []
        lead.message.technologies = []
        lead.message.locations = []
        lead.message.chat_history = []
        lead.chat_viewed_at = datetime.now(UTC)
        lead.chat_history = []
        lead.previous_publications = contact.previous_publications if hasattr(contact, "previous_publications") else []
        lead.bots = contact.bots if hasattr(contact, "bots") else []
        lead.contact = contact
        return lead


class UserTemplateModel(BaseModel):
    text: str
    subject: Optional[str]
    user_id: Optional[ObjectId]


class LinkedinContact(BaseModel):
    full_name: str
    slack_user: str
    title: str
    urls: List[dict]


class CloudFileModel(BaseModel):
    blob_path: str
    public_url: str
    file_name: str

    def __init__(self, blob_path: str, public_url: str, file_name: str):
        super().__init__()
        if not self.id:
            self.id = str(ObjectId())
        self.blob_path = blob_path
        self.public_url = public_url
        self.file_name = file_name


class GroupedMessagesModel:
    messages: List[ChatMessage] = []

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None

        model = cls()
        model.messages = [ChatMessage.from_dic(message) for message in dic.get('messages', [])]
        return model


class UserVerificationModel(DictionaryModel):
    pass

    def __init__(self):
        super().__init__()
        self.email = None
        self.created_at = datetime.now(UTC)


class UsersPage:
    users: List[UserModel]
    count: int = 0

    def __init__(self, users: List[UserModel], count: int):
        self.users = users
        self.count = count

    @staticmethod
    def from_dic(dic: dict):
        users = [UserModel.from_dic(doc) for doc in dic.get('page', [])]
        count = dic.get('count', 0)
        return UsersPage(users=users, count=count)


class Feature(DictionaryModel):
    display_name: str
    name: FeaturesEnum
    description: str | None = None
    limit: int | None = None
    options: FeatureOptions | None = None


class Subscription(BaseModel):
    features: list[Feature]
    duration_days: int
    name: str
    price: int
    limits: int
    trial: bool
    updated_at: datetime = None

    @classmethod
    def from_dic(cls, dic: dict):
        if not dic:
            return None
        model: Subscription | None = super().from_dic(dic)
        model.features = [Feature.from_dic(feature) for feature in dic.get('features', [])]
        return model



from enum import Enum, IntFlag


class DedicatedBotState(IntFlag):
    Operational = 0
    Stopped = 1
    ScheduledForDeletion = 2


class UserAccountState(IntFlag):
    Operational = 0
    Suspended = 1


class UserRole(str, Enum):
    ADMIN = 'admin'
    USER = 'user'
    MODERATOR = 'moderator'


class BotType(str, Enum):
    DEDICATED_BOT = 'dedicated',
    LEADGURU_BOT = 'leadguru'


class GoogleCloudFolder(str, Enum):
    SLACK_PROFILE_FILES = "Slack_profile"
    TICKET_FILES = "Ticket"


class SourceType(str, Enum):
    SLACK = 'slack'
    DISCORD = 'discord'


class BotUpdateEventType(str, Enum):
    NotDefined = 'NotDefined'
    BotAdded = 'BotAdded'
    BotDeleted = 'BotDeleted'
    BotUpdated = 'BotUpdated'


class UserAction(str, Enum):
    PAUSE_SOURCE = 'monitoring.pause.source'
    UNPAUSE_SOURCE = 'monitoring.unpause.source'
    STOP_CHANNEL = 'monitoring.stop.channel'
    STOP_SOURCE = 'monitoring.stop.source'
    START_CHANNEL = 'monitoring.start.channel'
    START_SOURCE = 'monitoring.start.source'
    LOGIN = 'login'
    LEAD_SAVE = 'lead.save'
    CHAT_MESSAGE = 'chat.message'
    ADMIN_CREDITS_ADDED = "admin-creds-added"
    ADMIN_CREDITS_SET = "admin-creds-set"
    INITIAL_CREDITS_SET = "initial-creds-set"


class StatusConnection(str, Enum):
    IN_PROGRESS = 'In progress',
    COMPLETE = 'Complete',
    FAILED = 'Failed'


class DefaultBoards(str, Enum):
    Inbox = 'Inbox',
    Primary = 'Primary board'


class BotEventType(str, Enum):
    CREATE = 'BotAdded'
    UPDATE = 'BotUpdated'
    DELETE = 'BotDeleted'


class ImageName(str, Enum):
    ARROW = 'arrow.png'
    FIREWORK = 'firework.png'
    LOCK = 'lock.png'
    LOGO = 'logo.png'
    MAIL = 'mail.png'
    CRY = 'cry.png'
    HANDS = 'hands.png'
    SAD = 'sad.png'


class FeaturesEnum(str, Enum):
    FEED = 'feed'
    PIPELINES = 'pipelines'
    LEADS = 'leads'
    CONTACTS = 'contacts'
    DASHBOARD = 'dashboard'
    EXPORT = 'export'
    CRM = 'crm'
    MESSAGE_REQUEST = 'message_request'
    DIRECT_MESSAGES = 'dm'
    PEOPLE = 'people'
    TEMPLATES = 'templates'
    BILLING = 'billing'
    PIPEDRIVE = 'pipedrive'
    MASS_DIRECT_MESSAGES = 'mass_direct_messages'
    TARGETS = 'targets'
    MY_TEAM = 'my_team'
    SCHEDULED_MESSAGES = 'scheduled_messages'
    SOURCES = 'sources'
    ALGORITHMS = 'algorithms'


class FeatureOptions(str, Enum):
    BASIC = 'basic'
    ADVANCED = 'advanced'



from abc import abstractmethod, ABC
from pydantic import BaseModel


class InvalidJobTypeException(Exception):
    pass


class BaseBackgroundJobData(BaseModel):
    """
    Background job data
    """


class BaseBackgroundJob(ABC):
    def __init__(self):
        super().__init__()

    @staticmethod
    def dumps(job_type: type, data: dict) -> dict:
        """
        @:param job_type type of the job, inherited from BaseBackgroundJob
        @:param job data
        """
        obj = {
            "job_type": job_type.__name__,
            "data": data
        }
        return obj

    @property
    @abstractmethod
    def job_data_type(self) -> type:
        """
       gets data type for this given job
       """

    @abstractmethod
    def exec(self, data: BaseBackgroundJobData):
        """
       @param data: job data
       """

    def run(self, data: dict):
        """
        @param data: job data
        """
        data_type = self.job_data_type
        if not issubclass(data_type, BaseBackgroundJobData):
            raise InvalidJobTypeException(f'{data_type} is not a subsclass of BaseBackgroundJobData')

        job_data: BaseBackgroundJobData = data_type(**data)
        return self.exec(job_data)



import os
import re
from typing import List, Optional
import pymongo
from dateutil import tz
from pymongo import MongoClient, UpdateOne
from bson.objectid import ObjectId
from lgt_jobs.lgt_data.enums import SourceType
from lgt_jobs.lgt_data.model import (LeadModel, BaseModel, UserModel, UserResetPasswordModel, BoardModel, BoardedStatus,
                                     DedicatedBotModel, SlackMemberInformation, UserTemplateModel, LinkedinContact,
                                     ExtendedUserLeadModel, UserLeadModel, ExtendedLeadModel, UserContact, ChatMessage,
                                     GroupedMessagesModel, UserVerificationModel, UsersPage, Subscription)
from datetime import datetime, UTC, timedelta
from collections import OrderedDict

client = MongoClient(os.environ.get('MONGO_CONNECTION_STRING', 'mongodb://127.0.0.1:27017/'))


def to_object_id(oid):
    if isinstance(oid, ObjectId):
        return oid
    return ObjectId(oid)


class BaseMongoRepository:
    collection_name = ''
    database_name = 'lgt_admin'
    model: BaseModel = BaseModel()

    def collection(self):
        return client[self.database_name][self.collection_name]

    def _collection(self, collection_name):
        return client[self.database_name][collection_name]

    def insert_many(self, items):
        insert_items = map(lambda x: x.to_dic(), items)
        self.collection().insert_many(insert_items)

    def insert(self, item: BaseModel):
        return self.collection().insert_one(item.to_dic())

    def add(self, item: BaseModel):
        return self.insert(item)

    def delete(self, id):
        res = self.collection().delete_one({'_id': to_object_id(id)})
        return res


class UserMongoRepository(BaseMongoRepository):
    collection_name = 'users'
    model = UserModel

    def get(self, _id):
        return UserModel.from_dic(self.collection().find_one({'_id': to_object_id(_id)}))

    def get_by_email(self, email: str):
        pipeline = {'email': email}
        doc = self.collection().find_one(pipeline)
        return UserModel.from_dic(doc)

    def set(self, _id, **kwargs):
        update_dict = {k: v for k, v in kwargs.items() if v is not None}
        self.collection().update_one({'_id': to_object_id(_id)}, {'$set': update_dict})

    def get_users(self, users_ids=None, include_inactive=False):
        pipeline = {}

        if users_ids is not None:
            pipeline['_id'] = {'$in': [to_object_id(_id) for _id in users_ids]}

        if not include_inactive:
            pipeline['inactive'] = False

        return [UserModel.from_dic(doc) for doc in self.collection().find(pipeline)]

    def get_users_page(self, skip: int = 0, limit: int = 50, **kwargs):
        pipeline = []
        match_pipeline = {}
        include_inactive = kwargs.get('include_inactive', False)
        email = kwargs.get('email')
        search = kwargs.get('search')
        sort_field = kwargs.get('sort_field', 'email')
        sort_direction = kwargs.get('sort_direction', 'ASCENDING')
        sort_direction: int = pymongo.ASCENDING if sort_direction == 'ASCENDING' else pymongo.DESCENDING
        if not include_inactive:
            match_pipeline['inactive'] = False
        if email:
            match_pipeline['email'] = email
        if search:
            match_pipeline['email'] = {'$regex': search, '$options': 'i'}
        pipeline.append({'$match': match_pipeline})
        pipeline.extend([
            {
                '$lookup': {
                    'from': 'user_track_action',
                    'as': 'user_actions',
                    'let': {'user_id': '$_id'},
                    'pipeline': [
                        {
                            '$match': {'$expr': {'$eq': ['$user_id', '$$user_id']}}
                        },
                        {
                            '$sort': {'created_at': -1}
                        },
                        {
                            '$limit': 1
                        }
                    ]
                }
            },
            {
                '$lookup': {
                    'from': 'subscriptions',
                    'localField': 'subscription_id',
                    'foreignField': '_id',
                    'as': 'subscription'
                }
            },
            {
                '$lookup': {
                    'from': 'subscriptions',
                    'as': 'expired_subscription',
                    'pipeline': [{'$match': {'expired': True}}]
                }
            },
            {
                '$addFields': {
                    'last_action_at': {'$first': '$user_actions.created_at'},
                    'subscription_name': {
                        '$cond': [
                            {'$gt': ['$subscription_expired_at', datetime.now(UTC)]},
                            {'$first': '$subscription.name'},
                            {'$first': '$expired_subscription.name'}
                        ]
                    },
                    'balance': {'$subtract': ['$leads_limit', '$leads_proceeded']},
                    'roles_length': {'$size': '$roles'}
                }
            },
            {
                '$unset': ['user_actions', 'slack_users', 'discord_users']
            }
        ])
        pipeline.append({'$sort': {sort_field: sort_direction}})
        pipeline.extend([{'$facet': {'page': [{'$skip': skip}, {'$limit': limit}], 'count': [{'$count': 'count'}]}},
                         {'$project': {'page': 1, 'count': {'$first': "$count.count"}}}])
        return UsersPage.from_dic(list(self.collection().aggregate(pipeline))[0])


class UserLeadMongoRepository(BaseMongoRepository):
    collection_name = 'user_leads'
    model = ExtendedUserLeadModel

    def get_count(self, user_id, **kwargs):
        pipeline = self.__create_leads_filter(user_id, **kwargs)
        return self.collection().count_documents(pipeline)

    def get_many(self, ids: list, user_id):
        docs = self.collection().find({"id": {'$in': ids}, 'user_id': to_object_id(user_id)})
        leads = [ExtendedUserLeadModel.from_dic(lead) for lead in docs]
        senders = [lead.message.sender_id for lead in leads]
        contacts = SlackContactUserRepository().find(user_id, users=senders)
        for lead in leads:
            lead.contact = next(filter(lambda x: x.sender_id == lead.message.sender_id, contacts), None)

        return leads

    def get_leads(self, user_id, skip: int, limit: int, **kwargs) -> List[ExtendedUserLeadModel]:
        pipeline = self.__create_leads_filter(user_id, **kwargs)
        sort_field = kwargs.get('sort_field', 'last_action_at')
        sort_direction = kwargs.get('sort_direction', 'ASCENDING')
        sort_direction = pymongo.ASCENDING if sort_direction == 'ASCENDING' else pymongo.DESCENDING
        docs = list(self.collection().find(pipeline).sort([(sort_field, sort_direction)]).skip(skip).limit(limit))
        leads = [ExtendedUserLeadModel.from_dic(x) for x in docs]
        senders = [lead.message.sender_id for lead in leads]
        contacts = SlackContactUserRepository().find(users=senders)
        for lead in leads:
            lead.contact = next(filter(lambda x: x.sender_id == lead.message.sender_id, contacts), None)
        return leads

    @staticmethod
    def __create_leads_filter(user_id, **kwargs):
        pipeline: dict = {'user_id': to_object_id(user_id)}

        if kwargs.get('status') is not None:
            pipeline['status'] = kwargs.get('status', '')

        if kwargs.get('board_id'):
            pipeline['board_id'] = to_object_id(kwargs.get('board_id'))
        elif kwargs.get('board_id') is not None:
            pipeline['$or'] = [{'board_id': ''}, {'board_id': None}]

        archived = kwargs.get('archived')
        from_date = kwargs.get('from_date')
        to_date = kwargs.get('to_date')
        has_followup = kwargs.get('has_followup', None)
        followup_to = kwargs.get('followup_to_date', None)
        followup_from = kwargs.get('followup_from_date', None)
        created_to = kwargs.get('created_to_date', None)
        created_from = kwargs.get('created_from_date', None)
        sender_ids = kwargs.get('sender_ids', None)
        text = kwargs.get('text', None)
        stop_words = kwargs.get('stop_words', None)
        tags = kwargs.get('tags', None)
        configs = kwargs.get('config', None)
        bots_names = kwargs.get('bots_names', None)
        locations = kwargs.get('locations', None)
        dedicated_bots_ids = kwargs.get('dedicated_bots_ids', None)
        with_chat = kwargs.get('with_chat', None)
        leads_ids = kwargs.get('leads_ids', None)
        exclude_leads = kwargs.get('exclude_leads', None)
        exclude_senders = kwargs.get('exclude_senders', None)

        pipeline['message.profile.display_name'] = {
            "$ne": "Slackbot"
        }

        if leads_ids is not None:
            pipeline["id"] = {'$in': leads_ids}

        if exclude_leads:
            pipeline['id'] = {'$nin': exclude_leads}

        if exclude_senders:
            pipeline['message.sender_id'] = {'$nin': exclude_senders}

        if archived is not None:
            pipeline['archived'] = archived

        if with_chat is not None:
            pipeline['chat_history'] = {'$exists': True, '$ne': []}

        if has_followup is not None:
            pipeline['followup_date'] = {'$ne': None} if has_followup else {'$eq': None}

        if from_date or to_date:
            pipeline['last_action_at'] = {}

        if from_date:
            start = datetime(from_date.year, from_date.month, from_date.day, tzinfo=tz.tzutc())
            pipeline['last_action_at']['$gte'] = start

        if to_date:
            end = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, tzinfo=tz.tzutc())
            pipeline['last_action_at']['$lte'] = end

        if locations and len(locations) > 0:
            pipeline['message.locations'] = {"$in": locations}

        if sender_ids:
            pipeline['message.sender_id'] = {'$in': sender_ids}

        if followup_from or followup_to:
            pipeline['followup_date'] = {}

        if followup_from:
            followup_from = datetime(followup_from.year, followup_from.month, followup_from.day, tzinfo=tz.tzutc())
            pipeline['followup_date']['$gte'] = followup_from

        if followup_to:
            followup_to = datetime(followup_to.year, followup_to.month, followup_to.day, 23, 59, 59, tzinfo=tz.tzutc())
            pipeline['followup_date']['$lte'] = followup_to

        if created_to or created_from:
            pipeline['created_at'] = {}

        if created_to:
            created_to = datetime(created_to.year, created_to.month, created_to.day, 23, 59, 59, tzinfo=tz.tzutc())
            pipeline['created_at']['$lte'] = created_to

        if created_from:
            created_from = datetime(created_from.year, created_from.month, created_from.day, tzinfo=tz.tzutc())
            pipeline['created_at']['$gte'] = created_from

        if stop_words:
            pipeline['full_message_text'] = {'$regex': f'^(?!.*({stop_words})).*$', '$options': 'i'}
        elif text:
            pipeline['$text'] = {'$search': text}

        if tags:
            pipeline["tags"] = {"$elemMatch": {"$in": tags}}

        if configs:
            pipeline["message.configs.id"] = {"$in": configs}

        if bots_names is not None:
            pipeline['message.name'] = {'$in': bots_names}

        if dedicated_bots_ids is not None:
            pipeline["message.dedicated_slack_options.bot_id"] = {"$in": dedicated_bots_ids}
        return pipeline

    def get_unanswered_leads_ids(self, user_id: str, from_date=None):
        pipeline = [
            {
                '$match':
                    {
                        'user_id': to_object_id(user_id)
                    }
            },
            {
                '$match':
                    {
                        'chat_history': {'$exists': True, '$not': {'$size': 0}}
                    }
            },
            {
                '$project':
                    {
                        'id': 1,
                        'sender_id': "$message.sender_id",
                        'last_message': {'$arrayElemAt': ['$chat_history', 0]}
                    }
            },
            {
                '$addFields':
                    {
                        'is_user_message': {'$ne': ['$sender_id', '$last_message.user']}
                    }
            },
            {
                '$match':
                    {
                        '$and':
                            [
                                {'last_message.text': {'$regex': '\\?'}},
                                {'is_user_message': False}
                            ]
                    }
            }
        ]

        if from_date:
            beginning_of_the_day = datetime(from_date.year, from_date.month, from_date.day, 0, 0, 0, 0)
            pipeline.insert(0, {"$match": {"created_at": {"$gte": beginning_of_the_day}}})

        data = list(self.collection().aggregate(pipeline))
        leads_ids = [item['id'] for item in data]
        return leads_ids

    def get_daily_analytics_by_workspace(self, user_configs: list,
                                         dedicated_only: bool | None,
                                         from_date: datetime,
                                         to_date: datetime,
                                         user_id: str):
        pipeline = [
            {
                '$addFields': {
                    'dedicated': {
                        '$anyElementTrue': {
                            '$map': {
                                'input': "$message.configs.id",
                                'as': "config",
                                'in': {'$in': ["$$config", user_configs]}
                            }
                        }
                    }
                }
            },
            {
                '$project': {
                    'created_at': {
                        '$dateToString': {
                            'format': '%Y-%m-%d',
                            'date': '$created_at'
                        }
                    },
                    'id': '$id'
                }
            },
            {
                '$group': {
                    '_id': '$created_at',
                    'data': {'$push': '$id'}
                }
            },
            {
                '$sort': {'_id': 1}
            }
        ]

        if dedicated_only:
            pipeline.insert(1, {"$match": {'dedicated': True}})
        elif dedicated_only is False:
            pipeline.insert(1, {"$match": {'dedicated': False}})

        if from_date:
            beginning_of_the_day = datetime(from_date.year, from_date.month, from_date.day, 0, 0, 0, 0)
            pipeline.insert(0, {"$match": {"created_at": {"$gte": beginning_of_the_day}}})

        if to_date:
            end_of_the_day = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, 999)
            pipeline.insert(0, {"$match": {"created_at": {"$lte": end_of_the_day}}})

        if user_id:
            pipeline.insert(0, {"$match": {'user_id': to_object_id(user_id)}})

        saved_messages = list(self.collection().aggregate(pipeline))
        saved_messages_dic = OrderedDict()

        for item in saved_messages:
            saved_messages_dic[item["_id"]] = item["data"]

        return saved_messages_dic

    def get_leads_after(self, action_after: datetime) -> [ExtendedUserLeadModel]:
        leads = self.collection().find({'last_action_at': {'$gte': action_after}})
        return [ExtendedUserLeadModel.from_dic(lead) for lead in leads]

    def add_lead(self, user_id, lead: UserLeadModel) -> None:
        if not lead.created_at:
            lead.created_at = datetime.now(UTC)

        if hasattr(lead, "_id"):
            lead._id = ObjectId()

        lead.user_id = user_id
        self.insert(lead)

    def update_lead(self, user_id, lead_id: str, **kwargs):
        pipeline = {'user_id': to_object_id(user_id), 'id': lead_id}
        update_dict = {k: v for k, v in kwargs.items() if v is not None}
        if 'board_id' in update_dict:
            update_dict['board_id'] = to_object_id(update_dict['board_id']) if len(update_dict['board_id']) == 24 \
                else update_dict['board_id']
        self.collection().update_one(pipeline, {'$set': update_dict})

        return ExtendedUserLeadModel.from_dic(self.collection().find_one(pipeline))

    def update_same_leads(self, user_id, sender_id: str, **kwargs):
        pipeline = {'user_id': to_object_id(user_id), 'message.sender_id': sender_id}
        update_dict = {k: v for k, v in kwargs.items() if v is not None}
        self.collection().update_many(pipeline, {'$set': update_dict})

    def update_leads_order(self, user_id, lead_ids: [str]):
        pipeline = {'user_id': to_object_id(user_id), 'id': {'$in': lead_ids}}
        docs = list(self.collection().find(pipeline))

        order = 0
        for lead_id in lead_ids:
            for doc in docs:
                if doc['id'] == lead_id:
                    self.collection().update_one({'id': lead_id}, {'$set': {'order': order}}, upsert=False)
                    order = order + 1

    def delete_lead(self, user_id, lead_id: str):
        """

        :param user_id:
        :param lead_id:
        :return: UserLeadModel
        """

        pipeline = {'user_id': to_object_id(user_id), 'id': lead_id}
        self.collection().delete_one(pipeline)

    def get_lead(self, user_id, message_id: str = None, lead_id: str = None, **kwargs):
        """

        :param user_id:
        :param message_id:
        :param lead_id:
        :return: UserLeadModel
        """

        pipeline = {'user_id': to_object_id(user_id)}
        sender_id = kwargs.get('sender_id')

        if message_id:
            pipeline['message.message_id'] = message_id

        if lead_id:
            pipeline['id'] = lead_id

        if sender_id:
            pipeline['message.sender_id'] = sender_id

        return ExtendedUserLeadModel.from_dic(self.collection().find_one(pipeline))


class UserResetPasswordMongoRepository(BaseMongoRepository):
    pass

    collection_name = 'user_reset_passwords'
    model = UserResetPasswordModel

    def get(self, _id):
        return UserResetPasswordModel.from_dic(self.collection().find_one({'_id': to_object_id(_id)}))

    def delete(self, email):
        self.collection().delete_many({'email': email})

    def add(self, email) -> str:
        model = UserResetPasswordModel()
        model.email = email
        return self.collection().insert_one({'email': email}).inserted_id


class LeadMongoRepository(BaseMongoRepository):
    pass

    database_name = 'lgt_admin'
    collection_name = 'general_leads'
    model = LeadModel

    def delete(self, _id):
        res = self.collection().delete_one({'id': _id})
        return res

    def get(self, _id=None, **kwargs):
        pipeline = {}
        timestamp = kwargs.get("timestamp")
        message_id = kwargs.get("message_id")
        channel_id = kwargs.get("channel_id")
        if _id:
            pipeline['id'] = _id
        if message_id:
            pipeline['message.message_id'] = message_id
        if channel_id:
            pipeline['message.channel_id'] = channel_id
        if timestamp:
            pipeline['message.timestamp'] = timestamp
        result: dict = self.collection().find_one(pipeline)
        if not result:
            return None

        return LeadModel.from_dic(result)

    def get_many(self, ids: list):
        docs = self.collection().find({"id": {'$in': ids}})
        leads = [ExtendedLeadModel.from_dic(lead) for lead in docs]
        senders = [lead.message.sender_id for lead in leads]
        contacts = SlackContactUserRepository().find(users=senders)
        for lead in leads:
            lead.contact = next(filter(lambda x: x.sender_id == lead.message.sender_id, contacts), None)

        return leads

    def get_by_sender_id(self, sender_id, exclude_leads: [str], skip: int, limit: int):
        pipeline = {'message.sender_id': sender_id, 'id': {'$nin': exclude_leads}}
        leads = self.collection().find(pipeline).sort([('created_at', pymongo.DESCENDING)]).skip(skip).limit(limit)
        return [LeadModel.from_dic(lead) for lead in leads]

    def get_by_message_id(self, message_id):
        """

        :rtype: LeadModel
        :param message_id:
        """
        doc: dict = self.collection().find_one({'message.message_id': message_id})
        if not doc:
            return None

        return LeadModel.from_dic(doc)

    def update(self, _id: str, **kwargs):
        update_dict = {k: v for k, v in kwargs.items() if v is not None}
        self.collection().update_one({'id': _id}, {'$set': update_dict}, upsert=False)

    def get_count(self, **kwargs):
        pipeline = self.__create_leads_filter(**kwargs)
        return self.collection().count_documents(pipeline)

    def get_list(self, skip, limit, **kwargs):
        pipeline = self.__create_leads_filter(**kwargs)

        sort_field = kwargs.get('sort_field', 'created_at')
        sort_direction = kwargs.get('sort_direction', 'ASCENDING')
        sort_direction = pymongo.ASCENDING if sort_direction == 'ASCENDING' else pymongo.DESCENDING

        leads = self.collection().find(pipeline).sort([(sort_field, sort_direction)]).skip(skip).limit(limit)
        return [LeadModel.from_dic(lead) for lead in leads]

    def __create_leads_filter(self, **kwargs):
        pipeline: dict = {"hidden": False}

        from_date: datetime = kwargs.get('from_date')
        to_date: datetime = kwargs.get('to_date')

        country = kwargs.get('country', None)
        user_id = kwargs.get('user_id', None)
        tags = kwargs.get('tags', None)
        text = kwargs.get('text', None)
        stop_words = kwargs.get('stop_words', None)
        exclude_leads = kwargs.get('exclude_leads', None)
        exclude_senders = kwargs.get('exclude_senders', None)
        excluded_channels = kwargs.get('excluded_channels', None)
        sender_ids = kwargs.get('sender_ids', None)
        configs = kwargs.get('config', None)
        bots_names = kwargs.get('bots_names', None)
        locations = kwargs.get('locations', None)
        publication_text = kwargs.get('publication_text')

        pipeline['message.profile.display_name'] = {
            "$ne": "Slackbot"
        }

        if from_date or to_date:
            pipeline['created_at'] = {}

        if from_date:
            start = from_date.astimezone(tz.tzutc())
            pipeline['created_at']['$gte'] = start

        if to_date:
            end = to_date.astimezone(tz.tzutc())
            pipeline['created_at']['$lte'] = end

        if locations and len(locations) > 0:
            pipeline['message.locations'] = {"$in": locations}

        if country:
            pipeline["message.slack_options.country"] = re.compile(country, re.IGNORECASE)

        if user_id:
            pipeline["$or"] = [
                {"message.dedicated_slack_options": {"$exists": False}},
                {"message.dedicated_slack_options": None},
                {"message.dedicated_slack_options.user_id": f"{user_id}"},
            ]
        else:
            pipeline["user_id"] = {'$exists': False}

        if stop_words:
            pipeline['full_message_text'] = {'$regex': f'^(?!.*({stop_words})).*$', '$options': 'i'}
        elif text:
            pipeline['$text'] = {'$search': text}

        if publication_text:
            pipeline['message.message'] = {'$regex': publication_text, '$options': 'i'}

        if tags:
            pipeline["tags"] = {"$elemMatch": {"$in": tags}}

        if exclude_leads:
            pipeline['id'] = {'$nin': exclude_leads}

        if exclude_senders:
            pipeline['message.sender_id'] = {'$nin': exclude_senders}

        if excluded_channels:
            pipeline['$and'] = []
            for ws, channels in excluded_channels.items():
                if channels is not None:
                    pipeline['$and'].append(
                        {'$or': [{'message.name': {'$ne': ws}}, {'message.channel_id': {'$nin': channels}}]})

        if sender_ids:
            pipeline['message.sender_id'] = {'$in': sender_ids}

        if configs:
            pipeline["message.configs.id"] = {"$not": {"$elemMatch": {"$nin": configs}}}

        if bots_names is not None:
            pipeline['message.name'] = {'$in': bots_names}

        pipeline['message.profile.real_name'] = {'$ne': 'Slackbot'}

        return pipeline

    def get_per_day(self, date: datetime):
        start_day = datetime(date.year, date.month, date.day, 0, 0, 0, tzinfo=tz.tzutc())
        end_day = datetime(date.year, date.month, date.day, 23, 59, 59, tzinfo=tz.tzutc())
        docs = self.collection().find({'created_at': {'$gte': start_day, '$lte': end_day}}).sort('created_at', 1)
        return [LeadModel.from_dic(x) for x in docs]


class SpamLeadsMongoRepository(LeadMongoRepository):
    pass

    def __init__(self):
        self.collection_name = 'spam_leads'

    def get_list(self, skip, limit, sort_field: str = 'created_at', sort_direction: str = 'ASCENDING', **kwargs):
        pipeline = self.__create_leads_filter(**kwargs)
        sort_direction = pymongo.ASCENDING if sort_direction == 'ASCENDING' else pymongo.DESCENDING
        docs = self.collection().find(pipeline).sort([(sort_field, sort_direction)]).skip(skip).limit(limit)
        leads = [ExtendedLeadModel.from_dic(doc) for doc in docs]
        senders = [lead.message.sender_id for lead in leads]
        contacts = SlackContactUserRepository().find(users=senders)
        for lead in leads:
            lead.contact = next(filter(lambda x: x.sender_id == lead.message.sender_id, contacts), None)
        return leads

    def get_count(self, **kwargs):
        pipeline = self.__create_leads_filter(**kwargs)
        return self.collection().count_documents(pipeline)

    def __create_leads_filter(self, **kwargs):
        pipeline = {"user_id": {'$exists': False}}
        text = kwargs.get('text', None)
        stop_words = kwargs.get('stop_words', None)
        if stop_words:
            pipeline['full_message_text'] = {'$regex': f'^(?!.*({stop_words})).*$', '$options': 'i'}
        elif text:
            pipeline['$text'] = {'$search': text}

        return pipeline


class GarbageLeadsMongoRepository(SpamLeadsMongoRepository):
    pass

    def __init__(self):
        super().__init__()
        self.collection_name = 'garbage_leads'


class GarbageUserLeadsMongoRepository(UserLeadMongoRepository):
    pass

    def __init__(self):
        self.database_name = 'lgt_admin'
        self.collection_name = 'garbage_leads'


class SpamUserLeadsMongoRepository(UserLeadMongoRepository):
    pass

    def __init__(self):
        self.database_name = 'lgt_admin'
        self.collection_name = 'spam_leads'


class BoardsMongoRepository(BaseMongoRepository):
    pass

    collection_name = 'boards'
    model = BoardModel

    def create_board(self, user_id: str, name: str, **kwargs):
        is_primary = kwargs.get('is_primary', False)

        if is_primary:
            primary_board: dict = self.collection().find_one({'user_id': to_object_id(user_id),
                                                              'is_primary': is_primary})
            if primary_board:
                return BoardModel.from_dic(primary_board)

        board = BoardModel()
        board.name = name
        board.created_at = datetime.now(UTC)
        board.user_id = to_object_id(user_id)
        board.is_primary = is_primary
        self.collection().insert_one(BoardModel.to_dic(board))

        return BoardModel.from_dic(self.collection().find_one({'user_id': to_object_id(user_id), 'name': name}))

    def add_default_statuses(self, user_id: str, board_id: str):
        pipeline = {'user_id': to_object_id(user_id), '_id': to_object_id(board_id)}
        board = BoardModel.from_dic(self.collection().find_one(pipeline))

        if not board:
            return None

        board.statuses.append(BoardedStatus().from_dic({'name': 'Lead', 'order': 0}))
        board.statuses.append(BoardedStatus().from_dic({'name': 'Prospect', 'order': 1}))
        board.statuses.append(BoardedStatus().from_dic({'name': 'Opportunity', 'order': 2}))
        board.statuses.append(BoardedStatus().from_dic({'name': 'Call', 'order': 3}))
        board.statuses.append(BoardedStatus().from_dic({'name': 'Contract', 'order': 4}))
        board.statuses.append(BoardedStatus().from_dic({'name': 'Refused', 'order': 5}))

        return self.update_board(user_id, board_id, statuses=board.statuses)

    def get(self, user_id: str, **kwargs):
        pipeline = {'user_id': to_object_id(user_id)}
        is_primary = kwargs.get('is_primary')
        default = kwargs.get('default')
        name = kwargs.get('name')

        if is_primary is not None:
            pipeline['is_primary'] = is_primary

        if default is not None:
            pipeline['default'] = default

        if name:
            pipeline['name'] = name

        docs = self.collection().find(pipeline).sort('created_at', 1)
        return [BoardModel.from_dic(doc) for doc in docs]

    def get_with_stats(self, user_id: str, **kwargs):
        is_primary = kwargs.get('is_primary')
        default = kwargs.get('default')
        name = kwargs.get('name')
        match = {'user_id': to_object_id(user_id)}

        if is_primary is not None:
            match['is_primary'] = is_primary

        if default is not None:
            match['default'] = default

        if name:
            match['name'] = name

        pipeline = [
            {
                '$match': match
            },
            {
                '$unwind': {'path': '$statuses', 'preserveNullAndEmptyArrays': True}
            },
            {
                '$lookup': {
                    'from': 'user_leads',
                    'as': 'statuses.user_leads',
                    'let': {'id': '$_id', 'status': '$statuses.name'},
                    'pipeline': [
                        {
                            '$match': {
                                '$expr': {
                                    '$and': [
                                        {'$eq': ['$board_id', '$$id']},
                                        {'$eq': ['$status', '$$status']},
                                        {'$eq': ['$archived', False]}
                                    ]
                                }
                            }
                        }
                    ]
                }
            },
            {
                '$addFields': {'statuses.user_leads': {'$size': '$statuses.user_leads'}}
            },
            {
                '$set': {'statuses': {'$cond': ['$statuses.name', '$statuses', None]}}
            },
            {
                '$group': {
                    '_id': '$_id',
                    'statuses': {'$push': '$statuses'},
                    'name': {'$first': '$name'},
                    'created_at': {'$first': '$created_at'},
                    'user_id': {'$first': '$user_id'},
                    'is_primary': {'$first': '$is_primary'},
                    'default': {'$first': '$default'}}
            },
            {
                '$addFields': {
                    'statuses': {'$filter': {'input': '$statuses', 'as': 'status', 'cond': {'$ne': ['$$status', None]}}}
                }
            },
            {'$sort': {'created_at': 1}}]

        return [BoardModel.from_dic(doc) for doc in self.collection().aggregate(pipeline)]

    def get_primary(self, user_id: str):
        return BoardModel.from_dic(self.collection().find_one({'user_id': to_object_id(user_id), 'is_primary': True}))

    def get_by_id(self, id: str):
        return BoardModel.from_dic(self.collection().find_one({'_id': to_object_id(id)}))

    def delete_by_id(self, id: str):
        return self.collection().delete_many({'_id': to_object_id(id)})

    def update_board(self, user_id, board_id: str, **kwargs):
        pipeline = {'user_id': to_object_id(user_id), '_id': to_object_id(board_id)}

        if kwargs.get('statuses'):
            kwargs['statuses'] = [status.to_dic() for status in kwargs.get('statuses')
                                  if isinstance(status, BoardedStatus)]

        doc = BoardModel.from_dic(self.collection().find_one(pipeline))
        if not doc:
            return None

        update_dict = {k: v for k, v in kwargs.items() if v is not None}
        self.collection().update_one(pipeline, {'$set': update_dict})

        self._collection('user_leads').update_many({'user_id': to_object_id(user_id), 'board_id': to_object_id(doc.id)},
                                                   {'$set': {'board_id': to_object_id(board_id)}})

        return BoardModel.from_dic(self.collection().find_one(pipeline))


class DedicatedBotRepository(BaseMongoRepository):
    pass

    collection_name = 'dedicated_bots'

    def get_by_user_and_source_id(self, user_id: str, source_id: str) -> Optional[DedicatedBotModel]:
        doc = self.collection().find_one({"user_id": ObjectId(f"{user_id}"), "source.source_id": source_id})
        return DedicatedBotModel.from_dic(doc)

    def get_by_user_and_active_server_id(self, user_id: str, active_server_id: str) -> Optional[DedicatedBotModel]:
        doc = self.collection().find_one({"user_id": ObjectId(f"{user_id}"), "active_servers.id": active_server_id})
        return DedicatedBotModel.from_dic(doc)

    def add_or_update(self, bot: DedicatedBotModel):
        bot_dict = bot.to_dic()
        if '_id' in bot_dict:
            bot_dict.pop('_id')
        update_response = self.collection().update_one({"source.source_id": bot.source.source_id,
                                                        "user_name": bot.user_name,
                                                        "user_id": to_object_id(bot.user_id)},
                                                       {'$set': bot_dict}, upsert=True)
        bot.id = update_response.upserted_id if update_response.upserted_id else bot.id
        return bot

    def get_all(self, only_valid: bool = False, **kwargs) -> List[DedicatedBotModel]:
        kwargs["only_valid"] = only_valid
        pipeline = self.__create_bots_filter(**kwargs)
        docs = self.collection().find(pipeline)
        return [DedicatedBotModel.from_dic(doc) for doc in docs]

    def get_source_ids_for_user(self, user_id, source_type):
        docs = self.collection().find({
            'user_id': user_id,
            'source.source_type': source_type
        }, {
            '_id': 0,
            'source_id': '$source.source_id'
        })
        source_ids = [item['source_id'] for item in docs]
        return source_ids

    def get_one(self, **kwargs):
        pipeline = self.__create_bots_filter(**kwargs)
        return DedicatedBotModel.from_dic(self.collection().find_one(pipeline))

    def get_user_bots(self, user_id: str, only_valid: bool = False,
                      include_deleted: bool = False, include_paused: bool = False, **kwargs) -> List[DedicatedBotModel]:
        pipeline = {'user_id': to_object_id(user_id)}
        user_name = kwargs.get('user_name')

        if user_name:
            pipeline['user_name'] = user_name

        if not include_deleted:
            pipeline['deleted'] = False

        if not include_paused:
            pipeline['paused'] = False

        if only_valid:
            pipeline['invalid_creds'] = False

        return [DedicatedBotModel.from_dic(doc) for doc in self.collection().find(pipeline)]

    def delete(self, _id: str):
        self.collection().update_one({'_id': to_object_id(f"{_id}")}, {"$set": {"deleted": True}})

    @staticmethod
    def __create_bots_filter(**kwargs):
        pipeline = {}
        name = kwargs.get('name')
        source_type = kwargs.get('source_type', None)
        user_name = kwargs.get('user_name')
        only_valid = kwargs.get('only_valid')
        include_paused = kwargs.get('include_paused', False)
        include_deleted = kwargs.get('include_deleted', False)
        bot_id = kwargs.get('id')
        user_id = kwargs.get('user_id')
        source_id = kwargs.get('source_id')
        server_id = kwargs.get('server_id')
        active_server_id = kwargs.get('active_server_id')

        if bot_id:
            pipeline["_id"] = to_object_id(bot_id)

        if user_id:
            pipeline["user_id"] = to_object_id(user_id)

        if name:
            pipeline["name"] = name

        if source_type:
            pipeline["source.source_type"] = source_type

        if user_name:
            pipeline["user_name"] = user_name

        if source_id:
            pipeline["source.source_id"] = source_id

        if server_id:
            pipeline["servers"] = {'$elemMatch': {'id': server_id}}

        if active_server_id:
            pipeline["servers"] = {'$elemMatch': {'id': active_server_id, 'deleted': False}}

        if only_valid:
            pipeline['invalid_creds'] = False

        if not include_deleted:
            pipeline['deleted'] = False

        if not include_paused and source_type == SourceType.DISCORD:
            pipeline['paused'] = False

        return pipeline


class SlackContactUserRepository(BaseMongoRepository):
    collection_name = "slack_contact"
    model = SlackMemberInformation

    def find(self, text: Optional[str] = None, skip: int = 0, limit: int = 1000, **kwargs):
        pipeline = {}

        users = kwargs.pop("users")
        presence_updated_at = kwargs.pop("online_updated_at", None)
        if text:
            pipeline['$text'] = {'$search': text}

        if users:
            pipeline['sender_id'] = {'$in': users}

        if presence_updated_at:
            pipeline['online_updated_at'] = {'$gte': presence_updated_at}

        pipeline = {**pipeline, **kwargs}

        docs = self.collection().find(pipeline).sort([("real_name", pymongo.ASCENDING)]) \
            .skip(skip) \
            .limit(limit)
        return [SlackMemberInformation.from_dic(doc) for doc in docs]

    def find_one(self, user_id: str):
        pipeline = {"sender_id": user_id}
        return SlackMemberInformation.from_dic(self.collection().find_one(pipeline))

    def get_count_in_workspaces(self):
        pipeline = [
            {
                '$match': {
                    'user': {
                        '$ne': 'USLACKBOT'
                    }
                }
            }, {
                '$group': {
                    '_id': '$workspace',
                    'count': {
                        '$sum': 1
                    }
                }
            }
        ]
        docs = list(self.collection().aggregate(pipeline))
        return {doc['_id']: doc['count'] for doc in docs}


class UserTemplatesRepository(BaseMongoRepository):
    collection_name = "user_templates"
    model = UserTemplateModel

    def get_all(self, user_id: str):
        return [UserTemplateModel.from_dic(doc) for doc in self.collection().find({'user_id': to_object_id(user_id)})]

    def get(self, id: str):
        return UserTemplateModel.from_dic(self.collection().find_one({'_id': to_object_id(id)}))

    def create_or_update(self, template: UserTemplateModel):
        result = self.collection().update_one(
            {"_id": to_object_id(template.id)},
            {'$set': template.to_dic()},
            upsert=True)

        if result.upserted_id:
            template.id = result.upserted_id

        return template

    def delete_by_id(self, id: str):
        return self.collection().find_one_and_delete({'_id': to_object_id(id)})


class LinkedinContactRepository(BaseMongoRepository):
    collection_name = "linkedin_contact"
    model = LinkedinContact

    def find(self, **kwargs):
        docs = self.collection().find({**kwargs})
        return [LinkedinContact.from_dic(doc) for doc in docs]


class UserContactsRepository(BaseMongoRepository):
    collection_name = 'user_contacts'

    def find(self, user_id: str, **kwargs):
        pipeline = {'user_id': to_object_id(user_id)}

        users = kwargs.get('users')
        spam = kwargs.get('spam')
        with_chat_only = kwargs.get('with_chat_only', False)
        source_id = kwargs.get('source_id')

        if users:
            pipeline['sender_id'] = {'$in': users}

        if spam is not None:
            pipeline['spam'] = spam

        if source_id:
            pipeline['source_id'] = source_id

        if with_chat_only:
            pipeline['chat_id'] = {'$ne': None}

        docs = self.collection().find(pipeline)
        return [UserContact.from_dic(doc) for doc in docs]

    def find_one(self, user_id: str, **kwargs):
        pipeline = {'user_id': to_object_id(user_id)}

        sender_id = kwargs.get('sender_id')
        source_id = kwargs.get('source_id')
        if sender_id:
            pipeline['sender_id'] = sender_id
        if source_id:
            pipeline['source_id'] = source_id

        if contact_instance := self.collection().find_one(pipeline):
            return UserContact.from_dic(contact_instance)

    def update(self, user_id: str | ObjectId, sender_id: str, source_id: str, **kwargs):
        pipeline = {'user_id': to_object_id(user_id), 'sender_id': sender_id, 'source_id': source_id}
        update_dict = {k: v for k, v in kwargs.items() if v is not None}
        self.collection().update_one(pipeline, {'$set': update_dict}, upsert=False)

    def find_grouped_actual_contacts(self, user_id: ObjectId | str, **kwargs) -> dict[str, list[UserContact]]:
        spam = kwargs.get('spam')
        with_chat_only = kwargs.get('with_chat_only', False)
        match_pipeline = {'user_id': to_object_id(user_id), 'source_id': {'$ne': None}}
        if spam is not None:
            match_pipeline['spam'] = spam
        if with_chat_only:
            match_pipeline['chat_id'] = {'$ne': None}
        pipeline = [
            {
                '$lookup': {
                    'from': 'user_leads',
                    'as': 'user_leads',
                    'let': {
                        'sender_id': '$sender_id',
                        'user_id': '$user_id'
                    },
                    'pipeline': [
                        {
                            '$match': {'$expr': {'$and': [
                                {'$eq': ['$message.sender_id', '$$sender_id']},
                                {'$eq': ['$user_id', '$$user_id']}]}}
                        }
                    ]
                }
            },
            {
                '$match': {'user_leads.last_action_at': {'$gt': datetime.now(UTC) - timedelta(days=90)}}
            },
            {
                '$unset': 'user_leads'
            },
            {
                '$group': {'_id': '$source_id', 'contacts': {'$push': '$$ROOT'}}
            },
            {
                '$project': {'source_id': '$_id', 'contacts': 1, '_id': 0}
            }
        ]
        pipeline.insert(0, {'$match': match_pipeline})
        contacts_groups = {}
        for doc in list(self.collection().aggregate(pipeline)):
            contacts_groups[doc['source_id']] = [UserContact.from_dic(contact) for contact in doc['contacts']]
        return contacts_groups

    def find_actual_contacts_users(self, days: int = 90):
        pipeline = [
            {
                '$lookup': {
                    'from': 'user_leads',
                    'as': 'leads',
                    'let': {'user_id': '$user_id', 'sender_id': '$sender_id'},
                    'pipeline': [
                        {'$sort': {'last_action_at': -1}},
                        {
                            '$match': {
                                '$expr': {
                                    '$and': [
                                        {'$eq': ['$$sender_id', '$message.sender_id']},
                                        {'$eq': ['$$user_id', '$user_id']}
                                    ]
                                }
                            }
                        },
                        {'$limit': 1}
                    ]
                }
            },
            {
                '$match': {'leads.last_action_at': {'$gte': datetime.now(UTC) - timedelta(days)}}
            },
            {
                '$group': {'_id': '$user_id'}
            }
        ]
        user_ids = []
        for doc in list(self.collection().aggregate(pipeline)):
            user_ids.append(doc['_id'])
        return user_ids


class TeamRepository(BaseMongoRepository):
    collection_name = 'teams'

    def get_teammate(self, user_id: str, teammate_id: str):
        pipeline = [
            {
                "$match": {
                    "$or": [
                        {"author_id": to_object_id(user_id), "recipient_id": to_object_id(teammate_id)},
                        {"author_id": to_object_id(teammate_id), "recipient_id": to_object_id(user_id)}
                    ],
                    "status": "approved"
                }
            }
        ]
        return list(self.collection().aggregate(pipeline))


class ChatRepository(BaseMongoRepository):
    collection_name = 'chat_messages'

    def get_list(self, **kwargs):
        pipeline = {}
        text = kwargs.get('text')
        user_id = kwargs.get('user_id')
        bot_id = kwargs.get('bot_id')
        sender_id = kwargs.get('sender_id')
        from_date = kwargs.get('from_date')
        to_date = kwargs.get('to_date')

        if text:
            pipeline['text'] = {"$regex": text, "$options": 'i'}
        if bot_id:
            pipeline['bot_id'] = to_object_id(bot_id)
        if user_id:
            pipeline['user_id'] = to_object_id(user_id)
        if sender_id:
            pipeline['sender_id'] = sender_id

        if from_date or to_date:
            pipeline['last_action_at'] = {}

        if from_date:
            start = datetime(from_date.year, from_date.month, from_date.day, tzinfo=tz.tzutc())
            pipeline['last_action_at']['$gte'] = start

        if to_date:
            end = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, tzinfo=tz.tzutc())
            pipeline['last_action_at']['$lte'] = end

        return [ChatMessage.from_dic(msg)
                for msg in self.collection().find(pipeline).sort("id", pymongo.ASCENDING)]

    def get_grouped_messages(self, **kwargs):
        match_pipeline = {}
        pipeline = [
            {
                '$sort': {'created_at': 1}
            },
            {
                '$group': {'_id': '$bot_id', 'messages': {'$push': '$$ROOT'}}
            }
        ]

        from_date = kwargs.get('from_date')
        to_date = kwargs.get('to_date')
        user_id = kwargs.get('user_id')

        if from_date or to_date:
            match_pipeline['created_at'] = {}

        if from_date:
            start = datetime(from_date.year, from_date.month, from_date.day, tzinfo=tz.tzutc())
            match_pipeline['created_at']['$gte'] = start

        if to_date:
            end = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, tzinfo=tz.tzutc())
            match_pipeline['created_at']['$lte'] = end

        if user_id:
            match_pipeline['user_id'] = to_object_id(user_id)

        pipeline.insert(0, {'$match': match_pipeline})
        return [GroupedMessagesModel.from_dic(doc) for doc in list(self.collection().aggregate(pipeline))]

    def upsert_messages(self, messages: list[dict]):
        if messages:
            operations = [UpdateOne({"id": msg['id'], "user_id": msg['user_id']},
                                    {'$set': msg}, upsert=True) for msg in messages]
            self.collection().bulk_write(operations)

    def delete_message(self, user_id: str, _id: str):
        self.collection().delete_one({"id": _id, 'user_id': to_object_id(user_id)})


class UserVerificationMongoRepository(BaseMongoRepository):
    pass

    collection_name = 'user_verifications'
    model = UserVerificationModel

    def get(self, _id):
        return UserVerificationModel.from_dic(self.collection().find_one({'_id': to_object_id(_id)}))

    def delete(self, email):
        self.collection().delete_many({'email': email})

    def add(self, email) -> str:
        model = UserVerificationModel()
        model.email = email
        model.created_at = datetime.now(UTC)
        return self.collection().insert_one(model.to_dic()).inserted_id


class SubscriptionsRepository(BaseMongoRepository):
    pass

    collection_name = 'subscriptions'
    model = Subscription

    def find_one(self, **kwargs):
        pipeline = {}
        subscription_id = kwargs.get('id')
        trial = kwargs.get('trial')
        expired = kwargs.get('expired')
        if subscription_id:
            pipeline['_id'] = to_object_id(subscription_id)
        if trial is not None:
            pipeline['trial'] = trial
        if expired is not None:
            pipeline['expired'] = expired

        return Subscription.from_dic(self.collection().find_one(pipeline))



import enum
from typing import Optional
from pydantic import BaseModel, Extra


class TelegramCommandType(enum.Enum):
    AUTH_VERIFICATION_CODE = 1
    HARD_RESET = 1


class TelegramCommand(BaseModel, extra=Extra.ignore):
    bot_id: str
    type: int
    data: Optional[str]



class SlackMethods:
    profile_set = 'users.profile.set'
    profile_get = 'users.profile.get'
    profile_set_photo = 'users.setPhoto'
    set_sections = 'users.profile.setSections'
    users_list = 'users.list'
    users_info = 'users.info'
    users_get_presence = 'users.getPresence'
    list_users = "users.list"

    conversations_join = 'conversations.join'
    conversations_leave = 'conversations.leave'
    conversations_list = 'conversations.list'
    conversations_open = 'conversations.open'
    conversations_history = 'conversations.history'
    conversations_replies = 'conversations.replies'
    conversations_info = 'conversations.info'

    chat_delete = 'chat.delete'
    chat_update = 'chat.update'
    chat_post_message = 'chat.postMessage'
    chat_schedule_message = "chat.scheduleMessage"
    chat_attachments = "chat.unfurlLink"

    rtm_connect = 'rtm.connect'

    reactions_get = 'reactions.get'
    create_shared_invite = 'users.admin.createSharedInvite'
    send_invite_by_email = 'users.admin.inviteBulk'
    alternative_invite_by_email = 'users.inviteRequests.create'

    upload_file = 'files.upload'
    download_file = 'files.download'
    delete_file = 'files.delete'
    share_files = 'files.share'

    check_email = 'signup.checkEmail'
    confirm_email = 'signup.confirmEmail'

    confirm_code = 'signin.confirmCode'
    find_workspaces = 'signin.findWorkspaces'

    auth_test = 'auth.test'

    team_profile_get = 'team.profile.get'
    team_info = 'team.info'



from abc import ABC
from lgt_jobs.lgt_common.slack_client.slack_client import SlackClient
from lgt_jobs.lgt_data.enums import SourceType
from lgt_jobs.lgt_data.model import UserModel
from lgt_jobs.lgt_data.mongo_repository import UserMongoRepository, DedicatedBotRepository
from pydantic import BaseModel
from lgt_jobs.basejobs import BaseBackgroundJobData, BaseBackgroundJob


class UpdateExternalUserProfileJobData(BaseBackgroundJobData, BaseModel):
    user_id: str


class UpdateExternalUserProfileJob(BaseBackgroundJob, ABC):

    @property
    def job_data_type(self) -> type:
        return UpdateExternalUserProfileJobData

    def exec(self, data: UpdateExternalUserProfileJobData):
        user = UserMongoRepository().get(data.user_id)
        bots = DedicatedBotRepository().get_all(user_id=data.user_id, only_valid=True, include_deleted=False)
        for bot in bots:
            if bot.type == SourceType.SLACK:
                slack = SlackClient(bot.token, bot.cookies)
                UpdateExternalUserProfileJob.__update_slack_profile(user, slack)

    @staticmethod
    def __update_slack_profile(user: UserModel, slack: SlackClient):
        profile = slack.get_team_profile()
        title_section_id = None
        title_field_id = None
        skype_section_id = None
        for field_data in profile.get('profile', {}).get('fields', []):
            if field_data.get('field_name') == 'title':
                title_section_id = field_data.get('section_id')
                title_field_id = field_data.get('id')
                break
        for section_data in profile.get('profile', {}).get('sections', []):
            if section_data.get('label') == 'About me':
                skype_section_id = section_data.get('id')
                break

        slack.update_profile(user.slack_profile.to_dic())
        auth = slack.test_auth().json()
        user_id = auth.get('user_id')
        title_element_id = title_field_id.replace(title_field_id[:2], 'Pe')
        response = slack.update_section(user_id, title_section_id, title_element_id, user.slack_profile.title)
        sections = response['result']['data']['setProfileSection']['profileSections']
        elements = []
        for section in sections:
            if section['label'] == 'About me':
                elements = section['profileElements']
                break
        skype_element_id = None
        for element in elements:
            if element['label'] == 'Skype':
                skype_element_id = element['elementId']
                break
        slack.update_section(user_id, skype_section_id, skype_element_id, user.slack_profile.title)
        # try to update user photo
        if user.photo_url:
            slack.update_profile_photo(user.photo_url)



import json
import sys
import threading
import time
from lgt_common.pubsub.pubsubfactory import PubSubFactory
import logging as log
from lgt_jobs.env import project_id, background_jobs_topic, background_jobs_subscriber
from lgt_jobs.runner import BackgroundJobRunner
from lgt_jobs import jobs_map
import google.cloud.logging

lock = threading.Lock()


def run_background_job(data):
    try:
        log.info(f"[JOB]: {data} [START]")
        BackgroundJobRunner.run(jobs_map=jobs_map, data=data)
        log.info(f"[JOB]: {data} [FINISHED]")
    except Exception:
        raise


def run_background_job_with_lock(message):
    try:
        data = json.loads(message.data)
        with lock:
            run_background_job(data)
    except:
        import traceback
        log.error(f"[ERROR][JOB]: {message.data} [ERROR] {traceback.format_exception(*sys.exc_info())} ")
        traceback.print_exception(*sys.exc_info())
    finally:
        # accept message any way
        message.ack()


if __name__ == '__main__':
    client = google.cloud.logging.Client()
    client.setup_logging()
    factory = PubSubFactory(project_id)
    factory.create_topic_if_doesnt_exist(background_jobs_topic)
    factory.create_subscription_if_doesnt_exist(background_jobs_subscriber, background_jobs_topic,600)
    bot_subscription_path = factory.get_subscription_path(background_jobs_subscriber, background_jobs_topic)
    factory.subscriber.subscribe(bot_subscription_path, callback=run_background_job_with_lock)
    while True:
        time.sleep(1)



import requests
from requests import Response
from lgt_jobs.env import v3_server_host, v1_server_host, service_account_email, service_account_password


class BaseHttpClient:
    api_version: str = None
    token: str = None

    def _create_default_headers(self):
        return {'Authorization': f"Bearer {self.token}"}


class V3ServerClient(BaseHttpClient):
    api_version = 'api/v3'

    def __init__(self):
        self.token = BillingServerClient.login(service_account_email, service_account_password)

    def update_user_lead(self, message_id: str, email: str, slack_channel: str,
                         board_id: str = None, status: str = None, followup_date=None) -> dict:
        payload = {
            'message_id': message_id,
            'user_email': email
        }

        update_lead_payload = {
            'slack_channel': slack_channel,
            'board_id': board_id,
            'status': status,
            'followup_date': followup_date
        }
        headers = self._create_default_headers()
        return requests.post(f'{v3_server_host}/{self.api_version}/user_leads/update',
                             headers=headers, params=payload, json=update_lead_payload).json()

    def save_lead_from_contact(self, sender_id: str, source_id: str, email: str, message: str = None) -> Response:
        payload = {
            'sender_id': sender_id,
            'source_id': source_id,
            'user_email': email,
            'message': message,
            'inbox': True
        }
        headers = self._create_default_headers()
        return requests.post(f'{v3_server_host}/{self.api_version}/user_leads/create',
                             headers=headers, params=payload)


class BillingServerClient:
    api_version = 'api'

    @staticmethod
    def login(email: str, password: str) -> str:
        payload = {
            'email': email,
            'password': password
        }
        response = requests.post(f'{v1_server_host}/{BillingServerClient.api_version}/login', json=payload).json()
        return response['access_token']



import os
from typing import Optional, List, Dict, Union
from collections import OrderedDict
from datetime import timedelta, datetime
from bson import ObjectId
from dateutil import tz
from lgt_jobs.lgt_data.mongo_repository import to_object_id
from pymongo import MongoClient

client = MongoClient(os.environ.get('MONGO_CONNECTION_STRING', 'mongodb://127.0.0.1:27017/'))
db = client.lgt_admin


def _build_date_aggregated_analytics_pipeline(source_id=None, email=None, started_at: datetime = None,
                                              ended_at: datetime = None, bots_ids: [str] = None):
    pipeline = [
        {
            "$sort": {"created_at": 1}
        },
        {
            "$project": {
                "created_at": {"$dateToString": {"format": "%Y-%m-%d", "date": "$created_at"}},
                "name": "$name"
            }
        },
        {
            "$group":
                {
                    "_id": "$created_at",
                    "count": {"$sum": 1}
                }
        },
        {
            "$project": {
                "_id": {"$dateFromString": {"format": "%Y-%m-%d", "dateString": "$_id"}},
                "count": "$count"
            }
        },
        {
            "$sort": {"_id": 1}
        },
        {"$limit": 1000}
    ]

    if source_id:
        pipeline.insert(0, {"$match": {"source.source_id": source_id}})

    if email:
        pipeline.insert(0, {"$match": {"name": email}})

    if started_at:
        beginning_of_the_day = datetime(started_at.year, started_at.month, started_at.day, 0, 0, 0, 0)
        pipeline.insert(0, {"$match": {"created_at": {"$gte": beginning_of_the_day}}})

    if ended_at:
        end_of_the_day = datetime(ended_at.year, ended_at.month, ended_at.day, 23, 59, 59, 999)
        pipeline.insert(0, {"$match": {"created_at": {"$lte": end_of_the_day}}})

    if bots_ids is not None:
        pipeline.insert(0, {"$match": {'extra_ids': {'$in': bots_ids}}})

    return pipeline


def _create_result_dic(started_at: datetime = None, ended_at: datetime = None):
    analytics_dict = OrderedDict()

    if started_at and ended_at:
        days_range = range(0, (ended_at - started_at).days + 1)
        for day in days_range:
            cur_date = started_at + timedelta(days=day)
            str_date = f'{cur_date.year}-{cur_date.month:02d}-{cur_date.day:02d}'
            analytics_dict[str_date] = 0

    return analytics_dict


def _prepare_date_analytics_doc(doc, ordered_result_dict: Dict[str, int]):
    for item in doc:
        str_date = f'{item["_id"].year}-{item["_id"].month:02d}-{item["_id"].day:02d}'
        ordered_result_dict[str_date] = item["count"]
    return ordered_result_dict


def get_aggregated_user_leads(user_id: Union[ObjectId, str],
                              from_date: datetime,
                              to_date: datetime = None):
    pipeline = [
        {
            '$match': {
                'created_at': {'$gte': from_date}
            }
        }, {
            '$group': {
                '_id': {
                    '$dateFromParts': {
                        'day': {
                            '$dayOfMonth': '$created_at'
                        },
                        'month': {
                            '$month': '$created_at'
                        },
                        'year': {
                            '$year': '$created_at'
                        }
                    }
                },
                'count': {
                    '$sum': 1
                },
                'leads': {
                    '$push': {
                        'id': '$id',
                        'created_at': '$created_at'
                    }
                }
            }
        }, {
            '$sort': {
                '_id': 1
            }
        }
    ]

    if user_id:
        pipeline[0]["$match"]["user_id"] = to_object_id(user_id)

    if to_date:
        end = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, tzinfo=tz.tzutc())
        pipeline[0]["$match"]["created_at"]['$lte'] = end

    user_leads_data = list(client.lgt_admin.user_leads.aggregate(pipeline))
    result = _create_result_dic(from_date, to_date)

    for item in user_leads_data:
        str_date = f'{item["_id"].year}-{item["_id"].month:02d}-{item["_id"].day:02d}'
        result[str_date] = item["count"]
    return result


def get_register_users_analytics(from_date: datetime = None, to_date: datetime = None):
    pipeline = [
        {
            '$addFields': {
                'created_at_formatted': {
                    '$dateToString': {
                        'format': '%Y-%m-%d',
                        'date': '$created_at'
                    }
                }
            }
        }, {
            '$group': {
                '_id': '$created_at_formatted',
                'count': {
                    '$sum': 1
                }
            }
        }
    ]
    if from_date:
        beginning_of_the_day = datetime(from_date.year, from_date.month, from_date.day, 0, 0, 0, 0)
        pipeline.insert(0, {"$match": {"created_at": {"$gte": beginning_of_the_day}}})

    if to_date:
        end_of_the_day = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, 999)
        pipeline.insert(0, {"$match": {"created_at": {"$lte": end_of_the_day}}})

    users = list(client.lgt_admin.users.aggregate(pipeline))
    users_dic = OrderedDict()

    for item in users:
        users_dic[item["_id"]] = item["count"]

    return users_dic


def get_global_saved_leads_analytics(from_date: datetime = None, to_date: datetime = None):
    pipeline = [
        {
            '$addFields': {
                'created_at_formatted': {
                    '$dateToString': {
                        'format': '%Y-%m-%d',
                        'date': '$created_at'
                    }
                }
            }
        }, {
            '$group': {
                '_id': '$created_at_formatted',
                'count': {
                    '$sum': 1
                }
            }
        }
    ]

    if from_date:
        beginning_of_the_day = datetime(from_date.year, from_date.month, from_date.day, 0, 0, 0, 0)
        pipeline.insert(0, {"$match": {"created_at": {"$gte": beginning_of_the_day}}})

    if to_date:
        end_of_the_day = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, 999)
        pipeline.insert(0, {"$match": {"created_at": {"$lte": end_of_the_day}}})

    saved_leads = list(client.lgt_admin.user_leads.aggregate(pipeline))
    saved_leads_dic = OrderedDict()
    for item in saved_leads:
        saved_leads_dic[str(item["_id"])] = item["count"]
    return saved_leads_dic


def get_global_uniq_leads_analytics(from_date: datetime = None, to_date: datetime = None):
    pipeline = [
        {
            '$addFields': {
                'created_at_formatted': {
                    '$dateToString': {
                        'format': '%Y-%m-%d',
                        'date': '$created_at'
                    }
                }
            }
        }, {
            '$group': {
                '_id': {
                    'created_at': '$created_at_formatted',
                    'message': '$message.message_id'
                },
                'uniq_leads': {
                    '$addToSet': '$message.message_id'
                }
            }
        }, {
            '$group': {
                '_id': '$_id.created_at',
                'uniq_leads_count': {
                    '$sum': {
                        '$size': '$uniq_leads'
                    }
                }
            }
        }
    ]

    if from_date:
        beginning_of_the_day = datetime(from_date.year, from_date.month, from_date.day, 0, 0, 0, 0)
        pipeline.insert(0, {"$match": {"created_at": {"$gte": beginning_of_the_day}}})

    if to_date:
        end_of_the_day = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, 999)
        pipeline.insert(0, {"$match": {"created_at": {"$lte": end_of_the_day}}})

    uniq_leads = list(client.lgt_admin.user_leads.aggregate(pipeline))
    uniq_leads_dic = OrderedDict()

    for item in uniq_leads:
        uniq_leads_dic[item["_id"]] = item["uniq_leads_count"]
    return uniq_leads_dic


def get_bots_global_analytics(from_date: datetime = None, to_date: datetime = None):
    pipeline = [
        {
            '$addFields': {
                'created_at_formatted': {
                    '$dateToString': {
                        'format': '%Y-%m-%d',
                        'date': '$created_at'
                    }
                }
            }
        },
        {
            '$group': {
                '_id': {
                    'source_id': '$source.source_id',
                    'created_at': '$created_at_formatted'
                },
                'count': {
                    '$sum': 1
                }
            }
        }, {
            '$group': {
                '_id': '$_id.created_at',
                'bots': {
                    '$sum': '$count'
                },
                'sources': {
                    '$sum': {
                        '$cond': [
                            {
                                '$eq': [
                                    '$count', 1
                                ]
                            }, 1, 0
                        ]
                    }
                }
            }
        }
    ]
    if from_date:
        beginning_of_the_day = datetime(from_date.year, from_date.month, from_date.day, 0, 0, 0, 0)
        pipeline.insert(0, {"$match": {"created_at": {"$gte": beginning_of_the_day}}})

    if to_date:
        end_of_the_day = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, 999)
        pipeline.insert(0, {"$match": {"created_at": {"$lte": end_of_the_day}}})

    bots = list(client.lgt_admin.dedicated_bots.aggregate(pipeline))
    result = {}
    for bot in bots:
        result.update({bot['_id']: {'bots': bot['bots'], 'sources': bot['sources']}})
    return result


def get_date_aggregated_analytics(source_id=None, started_at: datetime = None,
                                  ended_at: datetime = None, bots_ids: [str] = None, configs: [str] = None):
    pipeline: list = _build_date_aggregated_analytics_pipeline(source_id=source_id, started_at=started_at,
                                                               ended_at=ended_at, bots_ids=bots_ids)

    received_messages = list(db[f'received_messages'].aggregate(pipeline))

    if configs is not None:
        pipeline.insert(0, {"$match": {"attributes": {'$in': configs}}})

    filtered_messages = list(db[f'filtered_messages'].aggregate(pipeline))

    received_messages_dic = _create_result_dic(started_at, ended_at)
    filtered_messages_dic = _create_result_dic(started_at, ended_at)

    return (_prepare_date_analytics_doc(received_messages, received_messages_dic),
            _prepare_date_analytics_doc(filtered_messages, filtered_messages_dic))


def get_leads_aggregated_analytics(from_date: datetime = None, to_date: datetime = None, user_id: str = None):
    pipeline = [
        {
            '$project': {'_id': 0, 'user_ids': 1}
        },
        {
            '$unwind': {'path': '$user_ids', 'preserveNullAndEmptyArrays': False}
        },
        {
            '$group': {'_id': '$user_ids', 'count': {'$sum': 1}}
        }
    ]
    if user_id:
        pipeline.insert(0, {"$match": {"user_ids": {"$in": [str(user_id)]}}})

    if from_date:
        beginning_of_the_day = datetime(from_date.year, from_date.month, from_date.day)
        pipeline.insert(0, {"$match": {"created_at": {"$gte": beginning_of_the_day}}})

    if to_date:
        end_of_the_day = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, 999)
        pipeline.insert(0, {"$match": {"created_at": {"$lte": end_of_the_day}}})

    filtered_messages = list(client.lgt_admin.filtered_messages.aggregate(pipeline))
    filtered_messages_dic = OrderedDict()

    for item in filtered_messages:
        filtered_messages_dic[str(item["_id"])] = item["count"]

    return filtered_messages_dic


def get_contacts_aggregated_analytics(from_date: datetime = None, to_date: datetime = None):
    pipeline = [
        {
            '$group': {
                '_id': {
                    'user_id': '$user_id',
                    'sender_id': '$message.sender_id'
                },
                'count': {
                    '$sum': 1
                }
            }
        }, {
            '$project': {
                'user_id': '$_id.user_id',
                'sender_id': '$_id.sender_id',
                'count': 1,
                '_id': 0
            }
        }, {
            '$group': {
                '_id': '$user_id',
                'contacts_count': {
                    '$sum': 1
                }
            }
        }
    ]

    if from_date:
        beginning_of_the_day = datetime(from_date.year, from_date.month, from_date.day, 0, 0, 0, 0)
        pipeline.insert(0, {"$match": {"created_at": {"$gte": beginning_of_the_day}}})

    if to_date:
        end_of_the_day = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, 999)
        pipeline.insert(0, {"$match": {"created_at": {"$lte": end_of_the_day}}})

    contact_saved = list(client.lgt_admin.user_leads.aggregate(pipeline))
    contact_saved_dic = OrderedDict()
    for item in contact_saved:
        contact_saved_dic[str(item["_id"])] = item["contacts_count"]

    return contact_saved_dic


def get_total_active_bots_global_analytics(from_date: datetime = None, to_date: datetime = None):
    active_bot_dates = []
    while from_date <= to_date:
        active_bot_dates.append(from_date)
        from_date += timedelta(days=1)
    pipeline = [
        {
            '$match': {
                'created_at': {
                    '$lte': to_date
                },
                'deleted': False
            }
        }, {
            '$addFields': {
                'created_at_formatted': {
                    '$dateToString': {
                        'format': '%Y-%m-%d',
                        'date': '$created_at'
                    }
                }
            }
        }, {
            '$group': {
                '_id': {
                    'source': '$source.source_id',
                    'created_at': '$created_at_formatted'
                },
                'count': {
                    '$sum': 1
                }
            }
        }, {
            '$sort': {
                '_id': 1
            }
        }
    ]
    bots = list(client.lgt_admin.dedicated_bots.aggregate(pipeline))
    response = {}
    count = 0
    name = ""
    for date in active_bot_dates:
        for bot in bots:
            bot_date = datetime.strptime(bot["_id"]["created_at"], "%Y-%m-%d")
            if bot_date <= date:
                if bot["_id"]["source"] == name:
                    continue
                else:
                    count += 1
            name = bot["_id"]["source"]
        response[str(date.date())] = count
        count = 0

    return response


def get_bots_aggregated_analytics(from_date: datetime = None,
                                  to_date: datetime = None,
                                  bot_ids: Optional[List[str]] = None,
                                  configs: Optional[List[str]] = None):
    pipeline = [
        {
            "$group": {
                "_id": "$source.source_id",
                "count": {"$sum": 1}
            }
        },
        {"$limit": 1000}
    ]

    if from_date:
        beginning_of_the_day = datetime(from_date.year, from_date.month, from_date.day, 0, 0, 0, 0)
        pipeline.insert(0, {"$match": {"created_at": {"$gte": beginning_of_the_day}}})

    if to_date:
        end_of_the_day = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, 999)
        pipeline.insert(0, {"$match": {"created_at": {"$lte": end_of_the_day}}})

    if bot_ids is not None:
        pipeline.insert(0, {"$match": {"extra_ids": {"$in": bot_ids}}})

    received_messages = list(db[f'received_messages'].aggregate(pipeline))

    if configs is not None:
        pipeline.insert(0, {"$match": {"attributes": {"$in": configs}}})

    filtered_messages = list(db[f'filtered_messages'].aggregate(pipeline))
    received_messages_dic = OrderedDict()
    filtered_messages_dic = OrderedDict()

    for item in received_messages:
        received_messages_dic[item["_id"]] = item["count"]

    for item in filtered_messages:
        filtered_messages_dic[item["_id"]] = item["count"]

    return received_messages_dic, filtered_messages_dic


def get_channel_aggregated_analytics(from_date: datetime = None, to_date: datetime = None,
                                     bot_id: Optional[str | ObjectId] = None):
    pipeline = [
        {
            '$match': {
                'extra_ids': {"$in": [str(bot_id)]}
            }
        },
        {
            "$group": {
                '_id': {
                    '$arrayElemAt': [
                        '$attributes', 0
                    ]
                },
                'count': {"$sum": 1}
            }
        },
        {
            '$limit': 1000
        }
    ]

    if from_date:
        beginning_of_the_day = datetime(from_date.year, from_date.month, from_date.day, 0, 0, 0, 0)
        pipeline.insert(0, {"$match": {"created_at": {"$gte": beginning_of_the_day}}})

    if to_date:
        end_of_the_day = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, 999)
        pipeline.insert(0, {"$match": {"created_at": {"$lte": end_of_the_day}}})

    received_messages = list(db[f'received_messages'].aggregate(pipeline))
    filtered_messages = list(db[f'filtered_messages'].aggregate(pipeline))

    received_messages_dic = OrderedDict()
    filtered_messages_dic = OrderedDict()

    for item in received_messages:
        received_messages_dic[item["_id"]] = item["count"]

    for item in filtered_messages:
        filtered_messages_dic[item["_id"]] = item["count"]

    return received_messages_dic, filtered_messages_dic


def get_users_aggregated_analytics(event_type: str = 'user-lead-extended',
                                   from_date: datetime = None,
                                   to_date: datetime = None,
                                   email: str = None):
    pipeline = [
        {
            "$group": {
                "_id": "$name",
                "count": {"$sum": 1}
            }
        },
        {"$limit": 1000}
    ]

    if email:
        pipeline.insert(0, {"$match": {"name": email}})

    if from_date:
        beginning_of_the_day = datetime(from_date.year, from_date.month, from_date.day, 0, 0, 0, 0)
        pipeline.insert(0, {"$match": {"created_at": {"$gte": beginning_of_the_day}}})

    if to_date:
        end_of_the_day = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, 999)
        pipeline.insert(0, {"$match": {"created_at": {"$lte": end_of_the_day}}})

    read_messages = list(db[event_type].aggregate(pipeline))
    read_messages_dic = OrderedDict()

    for item in read_messages:
        read_messages_dic[item["_id"]] = item["count"]

    return read_messages_dic


def get_user_date_aggregated_analytics(email=None, event_type: str = 'user-lead-extended',
                                       started_at: datetime = None,
                                       ended_at: datetime = None):
    pipeline = _build_date_aggregated_analytics_pipeline(email=email, started_at=started_at, ended_at=ended_at)

    messages = list(db[event_type].aggregate(pipeline))
    messages_dic = _create_result_dic(started_at, ended_at)

    for item in messages:
        str_date = f'{item["_id"].year}-{item["_id"].month:02d}-{item["_id"].day:02d}'
        messages_dic[str_date] = item["count"]

    return messages_dic


def get_user_read_count(lead_ids: [str]):
    pipeline = [
        {
            "$group":
                {
                    "_id": "$data",
                    "count": {"$sum": 1}
                }
        },
        {
            "$match": {
                "_id": {"$in": lead_ids}
            }
        }
    ]
    messages = list(db['user-lead-extended'].aggregate(pipeline))
    result = dict()

    for message in messages:
        result[message['_id']] = message['count']

    return result


def get_events_leads(email, event, from_date, to_date=None):
    pipeline = {
        'name': email,
        'created_at': {'$gte': from_date}
    }

    if to_date:
        end = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, tzinfo=tz.tzutc())
        pipeline['created_at']['$lte'] = end

    return list(db[event].find(pipeline))


def get_leads_aggregation(email, event, from_date, to_date=None):
    pipeline = get_event_pipeline(email, from_date, to_date)
    return list(db[event].aggregate(pipeline))


def get_event_pipeline(email, from_date, to_date=None):
    pipeline = [
        {
            "$match": {
                'name': email,
                'created_at': {'$gte': from_date}
            }
        },
        {
            '$group': {
                '_id': {
                    '$dateFromParts': {
                        'day': {
                            '$dayOfMonth': '$created_at'
                        },
                        'month': {
                            '$month': '$created_at'
                        },
                        'year': {
                            '$year': '$created_at'
                        }
                    }
                },
                'count': {
                    '$sum': 1
                }
            }
        },
        {
            "$sort": {"_id": 1}
        }
    ]

    if to_date:
        end = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, tzinfo=tz.tzutc())
        pipeline[0]["$match"]["created_at"]['$lte'] = end

    return pipeline


def get_total_analytic_followup_up_date(from_date, to_date=None):
    beginning_of_the_day = datetime(from_date.year, from_date.month, from_date.day, 0, 0, 0, 0)
    end_of_the_day = datetime(to_date.year, to_date.month, to_date.day, 23, 59, 59, 999)
    pipeline = [
        {
            "$match": {
                "followup_date": {
                    "$gte": beginning_of_the_day,
                    "$lte": end_of_the_day
                }
            }
        },
        {
            "$group": {
                "_id": "$_id",
                "count": {"$sum": 1}
            }
        }
    ]
    followup_analytic = list(client.lgt_admin.user_leads.aggregate(pipeline))

    followup_analytic_dic = {item["_id"]: item["count"] for item in followup_analytic}

    return followup_analytic_dic



import os
import time
from google.cloud import pubsub_v1
from loguru import logger

project_id = os.environ.get('PUBSUB_PROJECT_ID')


def publish_message2_pubsub(topic_name, message_json):
    def callback(message_future):
        try:
            pubsub_result = message_future.result()
            # When timeout is unspecified, the exception method waits indefinitely.
            if message_future.exception(timeout=60):
                logger.error(f'Publishing message on {topic_name} threw an Exception {message_future.exception()}.')
            else:
                logger.info('lgt-metric:lgt-slack-aggregator:pub-sub:message-sent')
                logger.info(pubsub_result)
        except Exception:
            logger.error(f'Error has happening during getting result of future message')

    attempt = 0
    while True:
        try:
            publisher = pubsub_v1.PublisherClient()
            topic_path = publisher.topic_path(project_id, topic_name)
            logger.info(f'Json: {message_json}')
            message = publisher.publish(topic_path, data=bytes(message_json, "utf8"))
            message.add_done_callback(callback)
            result = message.result()
            logger.info(f'Message has been sent {result}')
            return
        except:
            attempt = attempt + 1
            if attempt >= 3:
                raise
            time.sleep(3)



from abc import ABC
from typing import Optional
import logging as log
from lgt_jobs.lgt_common.slack_client.web_client import SlackWebClient
from pydantic import BaseModel
from lgt_jobs.basejobs import BaseBackgroundJobData, BaseBackgroundJob

"""
Send Slack Code
"""


class SendSlackEmailJobData(BaseBackgroundJobData, BaseModel):
    email: str
    user_agent: str
    locale: Optional[str]


class SendSlackEmailJob(BaseBackgroundJob, ABC):
    @property
    def job_data_type(self) -> type:
        return SendSlackEmailJobData

    def exec(self, data: SendSlackEmailJobData):
        client = SlackWebClient('')
        code_sent = client.confirm_email(data.email, data.user_agent, data.locale)
        if not code_sent:
            log.warning(f'Unable to confirm code due to error: {code_sent}')
            return



from abc import ABC
from pydantic import BaseModel
import logging as log
from lgt_jobs.basejobs import BaseBackgroundJobData, BaseBackgroundJob
from lgt_jobs.env import smtp_login, smtp_password, smtp_host
from lgt_jobs.lgt_data.enums import ImageName

"""
Send email
"""


class SendMailJobData(BaseBackgroundJobData, BaseModel):
    html: str
    subject: str
    recipient: str
    sender: str = "noreply@leadguru.co"
    images: list[ImageName] = []


class SendMailJob(BaseBackgroundJob, ABC):
    @property
    def job_data_type(self) -> type:
        return SendMailJobData

    def exec(self, data: SendMailJobData):
        from redmail import EmailSender
        email = EmailSender(host=smtp_host, port=587, username=smtp_login, password=smtp_password, use_starttls=True)
        body_image = {}
        for image in data.images:
            body_image[f'IMAGE_{ImageName(image.value).name}'] = f'lgt_jobs/assets/images/{image.value}'

        email.send(
            sender=f"Leadguru <{data.sender}>",
            receivers=[data.recipient],
            subject=data.subject,
            html=data.html,
            body_images=body_image
        )
        log.info('email message has been sent')



name = "lgt_jobs"

from .jobs.user_balance_update import UpdateUserBalanceJob, UpdateUserBalanceJobData
from .jobs.send_slack_message import SendSlackMessageJob, SendSlackMessageJobData
from .jobs.analytics import (TrackAnalyticsJob, TrackAnalyticsJobData)
from .jobs.bot_stats_update import (BotStatsUpdateJob, BotStatsUpdateJobData)
from .jobs.chat_history import (LoadChatHistoryJob, LoadChatHistoryJobData)
from .jobs.update_slack_profile import (UpdateExternalUserProfileJob, UpdateExternalUserProfileJobData)
from .jobs.mass_message import SendMassMessageSlackChannelJob, SendMassMessageSlackChannelJobData
from .basejobs import (BaseBackgroundJobData, BaseBackgroundJob, InvalidJobTypeException)
from .smtp import (SendMailJob, SendMailJobData)
from .runner import (BackgroundJobRunner)
from .simple_job import (SimpleTestJob, SimpleTestJobData)

jobs_map = {
    "SimpleTestJob": SimpleTestJob,
    "BotStatsUpdateJob": BotStatsUpdateJob,
    "SendMailJob": SendMailJob,
    "TrackAnalyticsJob": TrackAnalyticsJob,
    "LoadChatHistoryJob": LoadChatHistoryJob,
    "UpdateExternalUserProfileJob": UpdateExternalUserProfileJob,
    "SendSlackMessageJob": SendSlackMessageJob,
    "UpdateUserBalanceJob": UpdateUserBalanceJob,
    "SendMassMessageSlackChannelJob": SendMassMessageSlackChannelJob,
}
__all__ = [
    # Jobs
    SimpleTestJob,
    BotStatsUpdateJob,
    SendMailJob,
    SimpleTestJob,
    LoadChatHistoryJob,
    UpdateExternalUserProfileJob,
    TrackAnalyticsJob,
    SendSlackMessageJob,
    UpdateUserBalanceJob,
    SendMassMessageSlackChannelJob,

    # module classes
    BackgroundJobRunner,
    BaseBackgroundJobData,
    BaseBackgroundJob,
    InvalidJobTypeException,
    BotStatsUpdateJobData,
    SendMailJobData,
    SimpleTestJobData,
    LoadChatHistoryJobData,
    UpdateExternalUserProfileJobData,
    TrackAnalyticsJobData,
    SendSlackMessageJobData,
    UpdateUserBalanceJobData,
    SendMassMessageSlackChannelJobData,
    # mapping
    jobs_map
]



from google.cloud import pubsub_v1
from google.api_core.exceptions import GoogleAPICallError
from google.auth.transport.grpc import *


class PubSubFactory:

    def __init__(self, project_id):
        self.project_id = project_id
        self.publisher = pubsub_v1.PublisherClient()
        self.subscriber = pubsub_v1.SubscriberClient()

    def get_topic_path(self, topic_name):
        return self.publisher.api.topic_path(self.project_id, f'{topic_name}')

    def get_subscription_path(self, subscriber_name, topic_name):
        return self.subscriber.api.subscription_path(self.project_id, f'{topic_name}_{subscriber_name}')

    def create_topic_if_doesnt_exist(self, topic_name) -> pubsub_v1.types.pubsub_gapic_types.Topic:
        topic_path = self.get_topic_path(topic_name)
        try:
            return self.publisher.api.get_topic(topic=topic_path)
        except GoogleAPICallError as ex:
            if ex.grpc_status_code == grpc.StatusCode.NOT_FOUND:
                return self.publisher.api.create_topic(name=topic_path)
            else:
                raise

    def delete_topic(self, topic_name: str):
        topic_path = self.get_topic_path(topic_name)
        self.publisher.api.delete_topic(topic=topic_path)

    def delete_subscriber(self, subscriber_name: str, topic_name: str):
        subscription_path = self.get_subscription_path(subscriber_name, topic_name)
        self.subscriber.api.delete_subscription(subscription=subscription_path)

    def create_subscription_if_doesnt_exist(self, subscriber_name, topic_name,
                                            ack_deadline_seconds=60) -> pubsub_v1.types.pubsub_gapic_types.Subscription:
        subscription_path = self.get_subscription_path(subscriber_name, topic_name)
        self.create_topic_if_doesnt_exist(topic_name)

        try:
            return self.subscriber.api.get_subscription(subscription=subscription_path)
        except GoogleAPICallError as ex:
            if ex.grpc_status_code == grpc.StatusCode.NOT_FOUND:
                return self.subscriber.api.create_subscription(name=subscription_path,
                                                               topic=self.get_topic_path(topic_name),
                                                               push_config=None,
                                                               ack_deadline_seconds=ack_deadline_seconds)
            else:
                raise



import requests
import aiohttp
import asyncio
import websockets
import json
import io
from urllib import parse
from requests import Response
from websockets.client import WebSocketClientProtocol
from .methods import SlackMethods


class SlackClient:
    base_url = 'https://slack.com/api/'
    token: str
    cookies: dict
    socket: WebSocketClientProtocol
    headers: dict

    def __init__(self, token: str, cookies):
        self.token = token
        self.headers = {"Authorization": f"Bearer {self.token}"}
        if isinstance(cookies, list):
            self.cookies = {cookie['name']: cookie['value'] for cookie in cookies}
        else:
            self.cookies = cookies

    def join_channels(self, channels):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        tasks = asyncio.gather(*[self.join_channel_async(channel) for channel in channels])
        results = loop.run_until_complete(tasks)
        loop.close()
        return results

    def leave_channels(self, channels):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        tasks = asyncio.gather(*[self.leave_channel_async(channel) for channel in channels])
        results = loop.run_until_complete(tasks)
        loop.close()
        return results

    async def join_channel_async(self, channel):
        async with aiohttp.ClientSession() as session:
            url = f'{self.base_url}{SlackMethods.conversations_join}?{self.__channel_payload(channel)}'
            async with session.post(url=url, cookies=self.cookies) as response:
                return await response.json()

    async def leave_channel_async(self, channel):
        async with aiohttp.ClientSession() as session:
            url = f'{self.base_url}{SlackMethods.conversations_leave}?{self.__channel_payload(channel)}'
            async with session.post(url=url, cookies=self.cookies) as response:
                return await response.json()

    def upload_file(self, file, file_name):
        payload = {"content": file, "filename": file_name}
        return requests.post(f"{self.base_url}{SlackMethods.upload_file}", data=payload,
                             headers=self.headers, cookies=self.cookies).json()

    def download_file(self, file_url) -> Response:
        return requests.get(file_url, headers=self.headers, cookies=self.cookies)

    def delete_file(self, file_id: str):
        payload = {"file": file_id}
        return requests.post(f"{self.base_url}{SlackMethods.delete_file}", data=payload,
                             headers=self.headers, cookies=self.cookies).json()

    def share_files(self, files_ids: list, channel: str, text: str = None) -> dict:
        payload = {
            "files": ','.join(files_ids),
            "channel": channel,
        }
        if text:
            payload["blocks"] = json.dumps([{"type": "rich_text", "elements": [
                {"type": "rich_text_section", "elements": [{"type": "text", "text": text}]}]}])

        return requests.post(f"{self.base_url}{SlackMethods.share_files}", data=payload,
                             headers=self.headers, cookies=self.cookies).json()

    def get_profile(self, user_id: str = None):
        url = f'{self.base_url}{SlackMethods.profile_get}'
        payload = {}
        if user_id:
            payload['user'] = user_id
        return requests.post(url=url, cookies=self.cookies, headers=self.headers, json=payload).json()

    def get_team_profile(self):
        url = f'{self.base_url}{SlackMethods.team_profile_get}'
        return requests.post(url=url, cookies=self.cookies, headers=self.headers).json()

    def get_team_info(self):
        url = f'{self.base_url}{SlackMethods.team_info}'
        return requests.get(url=url, cookies=self.cookies, headers=self.headers).json()

    def update_profile(self, profile):
        url = f'{self.base_url}{SlackMethods.profile_set}'
        return requests.post(url=url, headers=self.headers, cookies=self.cookies, json={'profile': profile}).json()

    def update_section(self, user_id: str, section_id: str, element_id: str, text: str):
        url = f'{self.base_url}{SlackMethods.set_sections}'
        payload = {
            'token': self.token,
            'user': user_id,
            'section': section_id,
            'elements': json.dumps([{"element_id": element_id, "text": {"text": text}}]),
        }
        return requests.post(url=url, headers=self.headers, cookies=self.cookies, data=payload).json()

    def update_profile_photo(self, photo_url):
        url = f'{self.base_url}{SlackMethods.profile_set_photo}'
        with requests.get(photo_url) as img_resp:
            if img_resp.status_code != 200:
                raise Exception(f"Invalid url: {photo_url}")
            image = io.BytesIO(img_resp.content)

            files = {"image": image}
            headers = {"Authorization": f"Bearer {self.token}"}

            return requests.post(url=url, files=files, headers=headers, cookies=self.cookies, verify=False).json()

    def get_conversations_list(self):
        payload = {'types': ','.join(["public_channel", "private_channel"])}
        base_url = f'{self.base_url}{SlackMethods.conversations_list}?limit=1000'
        page = requests.post(url=base_url, cookies=self.cookies, data=payload, headers=self.headers).json()
        result = page.copy()
        if page["ok"]:
            while page.get("response_metadata", {}).get("next_cursor"):
                url = base_url + f'&cursor={page.get("response_metadata", {}).get("next_cursor")}'
                page = requests.post(url=url, cookies=self.cookies, data=payload, headers=self.headers).json()
                if page['ok']:
                    result['channels'].extend(page['channels'])
            result["channels"] = [ch for ch in result["channels"]
                                  if ch.get('is_channel')
                                  and not ch.get('is_archived')
                                  and not ch.get('is_frozen')]

        return result

    def get_im_list(self):
        url = f'{self.base_url}{SlackMethods.conversations_list}?{self.__conversation_list_payload(["im"])}'
        return requests.get(url=url, cookies=self.cookies).json()

    def im_open(self, user: str):
        url = f'{self.base_url}{SlackMethods.conversations_open}?{self.__im_open_payload(user)}'
        return requests.post(url=url, cookies=self.cookies).json()

    def delete_message(self, channel: str, id: str):
        url = f'{self.base_url}{SlackMethods.chat_delete}?{self.__delete_message_payload(channel, id)}'
        return requests.post(url=url, cookies=self.cookies).json()

    def conversations_info(self, channel: str):
        url = f'{self.base_url}{SlackMethods.conversations_info}?{self.__conversation_info_payload(channel)}'
        return requests.post(url=url, cookies=self.cookies).json()

    def update_message(self, channel: str, id: str, text: str, file_ids: str):
        url = f'{self.base_url}{SlackMethods.chat_update}?{self.__update_message_payload(channel, id, text, file_ids)}'
        return requests.post(url=url, cookies=self.cookies).json()

    def conversations_history(self, channel: str, id: str = None):
        url = f'{self.base_url}{SlackMethods.conversations_history}?{self.__channel_payload(channel, id)}'
        return requests.get(url=url, cookies=self.cookies).json()

    def conversations_replies(self, channel: str, id: str):
        url = f'{self.base_url}{SlackMethods.conversations_replies}?{self.__ts_payload(channel, id)}'
        return requests.get(url=url, cookies=self.cookies).json()

    def get_presense(self, user: str = None):
        url = f'{self.base_url}{SlackMethods.users_get_presence}?{self.__presense_payload(user)}'
        return requests.get(url=url, cookies=self.cookies).json()

    def post_message(self, channel: str, text: str):
        import uuid
        payload = {
            'channel': channel,
            'text': text,
            'client_msg_id': str(uuid.uuid4())
        }
        url = f'{self.base_url}{SlackMethods.chat_post_message}'
        return requests.post(url=url, cookies=self.cookies, headers=self.headers, data=payload).json()

    def post_message_schedule(self, channel: str, text: str, post_at: int):
        url = (f'{self.base_url}{SlackMethods.chat_schedule_message}?'
               f'{self.__post_scheduled_message_payload(channel, text, post_at)}')
        return requests.post(url=url, cookies=self.cookies).json()

    def users_list(self, cursor=None, limit: int = 1000):
        url = f'{self.base_url}{SlackMethods.users_list}?{self.__token_payload()}'
        if cursor:
            url += f'&cursor={cursor}'

        if limit:
            url += f'&limit={limit}'

        return requests.get(url=url, cookies=self.cookies).json()

    def user_info(self, user: str):
        url = f'{self.base_url}{SlackMethods.users_info}?{self.__user_info_payload(user)}'
        return requests.get(url=url, cookies=self.cookies).json()

    def get_reactions(self, channel: str, id: str):
        url = f'{self.base_url}{SlackMethods.reactions_get}?{self.__get_reactions_payload(channel, id)}'
        return requests.get(url=url, cookies=self.cookies).json()

    def get_attachments(self, channel: str, msg_id: str, url: str):
        payload = {
            'token': self.token,
            'channel': channel,
            'client_msg_id': msg_id,
            'url': url
        }
        url = f'{self.base_url}{SlackMethods.chat_attachments}'
        response = requests.post(url=url, cookies=self.cookies, data=payload)
        if response.status_code != 200:
            return
        return response.json().get('attachments')

    def check_email(self, email: str, user_agent: str) -> bool:
        payload = {'email': email}
        headers = {'User-Agent': user_agent}
        response = requests.post(f"{self.base_url}/{SlackMethods.check_email}", params=payload, headers=headers)
        if response.status_code != 200:
            return False
        return response.json()['ok']

    def confirm_email(self, email: str, user_agent: str, locale: str = 'en-US') -> bool:
        payload = {'email': email, 'locale': locale}
        headers = {'User-Agent': user_agent}
        response = requests.post(f"{self.base_url}/{SlackMethods.confirm_email}", params=payload, headers=headers)
        if response.status_code != 200:
            return False
        return response.json()['ok']

    def confirm_code(self, email: str, code: str, user_agent: str) -> requests.Response:
        payload = {'email': email, 'code': code}
        headers = {'User-Agent': user_agent}
        return requests.post(f"{self.base_url}/{SlackMethods.confirm_code}", params=payload, headers=headers)

    def find_workspaces(self, user_agent: str) -> requests.Response:
        headers = {'User-Agent': user_agent}
        response = requests.post(f"{self.base_url}/{SlackMethods.find_workspaces}",
                                 cookies=self.cookies, headers=headers)
        return response

    def create_shared_invite(self):
        expiration = '36000'
        max_signups = '100'
        payload = {
            'expiration': expiration,
            'max_signups': max_signups
        }
        headers = {"Authorization": f"Bearer {self.token}"}
        response = requests.post(f"{self.base_url}/{SlackMethods.create_shared_invite}", headers=headers,
                                 cookies=self.cookies, params=payload)
        return response

    def send_slack_invite_to_workspace(self, email):
        payload = {"invites": [{'email': email, "type": "regular", "mode": "manual"}]}
        headers = {"Authorization": f"Bearer {self.token}"}
        response = requests.post(f"{self.base_url}/{SlackMethods.send_invite_by_email}", headers=headers,
                                 cookies=self.cookies, json=payload)
        if not response.json()['ok']:
            payload = {"requests": payload["invites"]}
            response = requests.post(f"{self.base_url}/{SlackMethods.alternative_invite_by_email}", headers=headers,
                                     cookies=self.cookies, json=payload)
        return response

    def test_auth(self):
        headers = {"Authorization": f"Bearer {self.token}"}
        return requests.post(f"{self.base_url}/{SlackMethods.auth_test}", headers=headers,
                             cookies=self.cookies)

    def rtm_connect(self, callback):
        loop = asyncio.new_event_loop()
        loop.run_until_complete(self.__consumer(callback))
        loop.run_forever()

    async def __consumer(self, callback):
        url = f'{self.base_url}{SlackMethods.rtm_connect}?{self.__token_payload()}'
        response = requests.get(url=url, cookies=self.cookies).json()
        web_socket_url = response['url']
        async with websockets.connect(uri=web_socket_url) as websocket:
            self.socket = websocket
            async for message in websocket:
                await callback(json.loads(message))

    def __token_payload(self):
        return parse.urlencode({'token': self.token})

    def __user_info_payload(self, user):
        payload = {
            'token': self.token,
            'user': user
        }
        return parse.urlencode(payload)

    def __post_scheduled_message_payload(self, channel, text, post_at):
        payload = {
            'token': self.token,
            'channel': channel,
            'text': text,
            'post_at': post_at
        }
        return parse.urlencode(payload)

    def __update_message_payload(self, channel, id, text, file_ids):
        payload = {
            'parse': 'none',
            'token': self.token,
            'channel': channel,
            'ts': id,
            'text': text,
            'file_ids': file_ids
        }
        return parse.urlencode(payload)

    def __conversation_info_payload(self, channel):
        payload = {
            'token': self.token,
            'channel': channel,
            'include_num_members': "true"
        }
        return parse.urlencode(payload)

    def __delete_message_payload(self, channel, id):
        payload = {
            'token': self.token,
            'channel': channel,
            'ts': id
        }
        return parse.urlencode(payload)

    def __conversation_list_payload(self, types: list):
        payload = {
            'token': self.token,
            'types': ','.join(types)
        }
        return parse.urlencode(payload)

    def __im_open_payload(self, user: str):
        payload = {
            'token': self.token,
            'users': user,
            'types': 'im'
        }
        return parse.urlencode(payload)

    def __update_profile_payload(self, profile):
        payload = {
            'token': self.token,
            'profile': profile
        }
        return parse.urlencode(payload)

    def __channel_payload(self, channel, id=None):
        payload = {
            'token': self.token,
            'channel': channel
        }

        if id:
            payload["ts"] = id
            payload["limit"] = 1
            payload["inclusive"] = True

        return parse.urlencode(payload)

    def __presense_payload(self, user: str = None):
        payload = {
            'token': self.token,
        }

        if user:
            payload["user"] = user
        return parse.urlencode(payload)

    def __ts_payload(self, channel: str, id: str):
        payload = {
            'token': self.token,
            'channel': channel,
            'ts': id
        }
        return parse.urlencode(payload)

    def __get_reactions_payload(self, channel, id):
        payload = {
            'token': self.token,
            'full': True,
            'channel': channel,
            'timestamp': id
        }
        return parse.urlencode(payload)



from datetime import datetime, UTC
from abc import ABC
from typing import Optional
import logging as log
from lgt_jobs.lgt_common.slack_client.web_client import SlackWebClient, SlackMessageConvertService
from lgt_jobs.lgt_data.model import ChatMessage
from lgt_jobs.lgt_data.mongo_repository import DedicatedBotRepository, UserContactsRepository, UserMongoRepository, \
    ChatRepository
from pydantic import BaseModel
from lgt_jobs.basejobs import BaseBackgroundJobData, BaseBackgroundJob

"""
Send Slack Message
"""


class SendSlackMessageJobData(BaseBackgroundJobData, BaseModel):
    sender_id: str
    bot_id: str
    text: Optional[str]
    files_ids: Optional[list] = []


class SendSlackMessageJob(BaseBackgroundJob, ABC):
    @property
    def job_data_type(self) -> type:
        return SendSlackMessageJobData

    def exec(self, data: SendSlackMessageJobData):
        bot = DedicatedBotRepository().get_one(id=data.bot_id)
        if not bot:
            return

        user = UserMongoRepository().get(bot.user_id)
        user_contacts_repository = UserContactsRepository()
        uc = user_contacts_repository.find_one(bot.user_id, sender_id=data.sender_id,
                                               source_id=bot.source.source_id)
        uc.scheduled_messages = [message.to_dic() for message in uc.scheduled_messages
                                 if message.post_at.replace(tzinfo=UTC) > datetime.now(UTC)]
        user_contacts_repository.update(bot.user_id, uc.sender_id, bot.source.source_id,
                                        scheduled_messages=uc.scheduled_messages)
        slack_client = SlackWebClient(bot.token, bot.cookies)
        im = slack_client.im_open(data.sender_id)
        if not im['ok']:
            log.warning(f"Unable to open im with user: {data.sender_id}")
            return

        channel_id = im['channel']['id']
        if data.files_ids:
            resp = slack_client.share_files(data.files_ids, channel_id, data.text)
        else:
            resp = slack_client.post_message(channel_id, data.text)

        message = resp.get('message') if 'message' in resp \
            else slack_client.conversation_replies(channel_id, resp['file_msg_ts'])['messages'][0]

        message_model: ChatMessage = SlackMessageConvertService.from_slack_response(user.email, "slack_files",
                                                                                    bot.token, message, data.sender_id,
                                                                                    bot.id, user.id,
                                                                                    slack_client.client.cookies)
        message_model.viewed = True
        ChatRepository().upsert_messages([message_model.to_dic()])



import datetime
from abc import ABC
from typing import Optional
from lgt_jobs.lgt_data.engine import UserCreditStatementDocument
from lgt_jobs.lgt_data.enums import UserAccountState, UserAction
from lgt_jobs.lgt_data.model import UserModel
from lgt_jobs.lgt_data.mongo_repository import UserMongoRepository, to_object_id
from pydantic import BaseModel
from lgt_jobs.basejobs import BaseBackgroundJobData, BaseBackgroundJob

"""
User balance handling
"""


class UpdateUserBalanceJobData(BaseBackgroundJobData, BaseModel):
    user_id: Optional[str]


class UpdateUserBalanceJob(BaseBackgroundJob, ABC):
    @property
    def job_data_type(self) -> type:
        return UpdateUserBalanceJobData

    @staticmethod
    def __update_processed(user: UserModel):
        pipeline = [
            {
                "$match": {
                    "user_id": to_object_id(user.id),
                    "balance": {"$lte": 0},
                    "action": {"$nin": [UserAction.ADMIN_CREDITS_ADDED, UserAction.ADMIN_CREDITS_SET]}
                }
            },
            {
                '$group': {
                    '_id': "$user_id",
                    'count': {'$sum': {"$multiply": [-1, "$balance"]}}
                }
            }
        ]

        results = list(UserCreditStatementDocument.objects.aggregate(*pipeline))
        if not results:
            return

        count = results[0]["count"]
        if count >= user.leads_limit:
            # suspend account
            if user.state != UserAccountState.Suspended.value:
                UserMongoRepository().set(user.id, leads_proceeded=count,
                                          credits_exceeded_at=datetime.datetime.now(datetime.UTC),
                                          state=UserAccountState.Suspended.value)
                return

        if user.state == UserAccountState.Suspended.value:
            # unsuspend account
            UserMongoRepository().set(user.id, leads_proceeded=count,
                                      credits_exceeded_at=None,
                                      state=UserAccountState.Operational.value)
            return

        UserMongoRepository().set(user.id, leads_proceeded=count)

    def exec(self, data: UpdateUserBalanceJobData):
        user = UserMongoRepository().get(data.user_id)
        if not user:
            return

        self.__update_processed(user=user)



from abc import ABC
from typing import Optional
from pydantic import BaseModel
from .basejobs import BaseBackgroundJobData, BaseBackgroundJob


class SimpleTestJobData(BaseBackgroundJobData, BaseModel):
    id: int
    name: Optional[str] = None


class SimpleTestJob(BaseBackgroundJob, ABC):
    @property
    def job_data_type(self) -> type:
        return SimpleTestJobData

    def exec(self, data: SimpleTestJobData):
        print(f'id={data.id};name={data.name}')
        return f"id={data.id};name={data.name}"



from datetime import datetime, UTC
from abc import ABC
from typing import Optional, List
import logging as log
from lgt_jobs.lgt_common.slack_client.web_client import SlackWebClient, SlackMessageConvertService
from lgt_jobs.lgt_data.helpers import get_help_text
from lgt_jobs.lgt_data.model import ChatMessage, UserModel, UserContact, DedicatedBotModel
from lgt_jobs.lgt_data.mongo_repository import UserMongoRepository, DedicatedBotRepository, UserContactsRepository, \
    ChatRepository
from pydantic import BaseModel
from lgt_jobs.lgt_data.enums import SourceType, ImageName
from lgt_jobs.runner import BaseBackgroundJob, BaseBackgroundJobData, BackgroundJobRunner
from lgt_jobs.env import portal_url
from lgt_jobs.smtp import SendMailJobData, SendMailJob

"""
Load slack chat history
"""


class LoadChatHistoryJobData(BaseBackgroundJobData, BaseModel):
    user_id: str
    template_path: str = 'lgt_jobs/templates/new_message.html'


class LoadChatHistoryJob(BaseBackgroundJob, ABC):
    chat_repo = ChatRepository()
    contacts_repo = UserContactsRepository()

    @property
    def job_data_type(self) -> type:
        return LoadChatHistoryJobData

    def exec(self, data: LoadChatHistoryJobData):
        user = UserMongoRepository().get(data.user_id)
        if user.subscription_expired_at.replace(tzinfo=UTC) < datetime.now(UTC):
            log.info(f"[LoadChatHistoryJob]: {user.email} has expired subscription")
            return
        bots = DedicatedBotRepository().get_all(only_valid=True, user_id=user.id, source_type=SourceType.SLACK)
        if not bots:
            return
        last_message = None
        last_message_contact = None
        last_message_bot = None
        contacts_groups = self.contacts_repo.find_grouped_actual_contacts(user.id, spam=False, with_chat_only=False)
        for bot in bots:
            contacts = contacts_groups.get(bot.source.source_id)
            if not contacts:
                continue

            log.info(f"[LoadChatHistoryJob]: processing {len(contacts)} contacts for user: {user.email}")
            for contact in contacts:
                message = self._update_history(user=user, contact=contact, bot=bot)

                if not message:
                    continue

                if not last_message:
                    last_message_bot = bot
                    last_message = message
                    last_message_contact = contact

                if message.created_at > last_message.created_at and message.user == contact.sender_id:
                    last_message_bot = bot
                    last_message = message
                    last_message_contact = contact

        has_to_be_notified = (not user.new_message_notified_at or
                              (last_message and last_message.created_at > user.new_message_notified_at))

        if last_message and has_to_be_notified and last_message.user == last_message_contact.sender_id:
            LoadChatHistoryJob._notify_about_new_messages(user, last_message_contact, last_message_bot,
                                                          data.template_path)
            UserMongoRepository().set(data.user_id, new_message_notified_at=datetime.now(UTC))

    def _get_new_messages(self, contact: UserContact, bot: DedicatedBotModel, slack_chat: List[ChatMessage]):
        messages = self.chat_repo.get_list(sender_id=contact.sender_id, bot_id=bot.id)
        new_messages = []
        for message in slack_chat:
            same_messages = [msg for msg in messages if msg.id == message.id]
            if not same_messages:
                new_messages.append(message)
        return new_messages

    def _update_history(self, user: UserModel, contact: UserContact, bot: DedicatedBotModel) -> Optional[ChatMessage]:
        slack_client = SlackWebClient(bot.token, bot.cookies)
        try:
            chat_id = slack_client.im_open(contact.sender_id).get('channel', {}).get('id')
            history = slack_client.chat_history(chat_id)
        except Exception as ex:
            log.error(f'[LoadChatHistoryJob]: Failed to load chat for the contact: {contact.id}. ERROR: {str(ex)}')
            return

        if not history['ok']:
            log.error(f'[LoadChatHistoryJob]: Failed to load chat for the contact: {contact.id}. '
                      f'ERROR: {history.get("error", "")}')
            return

        messages = history.get('messages', [])
        if not messages:
            return None

        messages = [SlackMessageConvertService.from_slack_response(user.email, "slack_files",
                                                                   bot.token, m, contact.sender_id, bot.id,
                                                                   user.id, bot.cookies) for m in messages]
        new_messages = self._get_new_messages(contact, bot, messages)
        chat_history = [message.to_dic() for message in new_messages]
        self.chat_repo.upsert_messages(chat_history)
        if bot.associated_user != contact.sender_id and new_messages:
            log.info(f'[LoadChatHistoryJob]: New message. Sender id: {contact.sender_id}, bot id: {bot.id}')
            self.contacts_repo.update(contact.user_id, contact.sender_id, contact.source_id,
                                      last_message_at=datetime.now(UTC))
            return new_messages[-1]

        return None

    @staticmethod
    def _notify_about_new_messages(user: UserModel, contact: UserContact, bot: DedicatedBotModel, template_path: str):
        with open(template_path, mode='r') as template_file:
            html = template_file.read()
            chat_url = f'{portal_url}/feed?senderId={contact.sender_id}&sourceId={bot.source.source_id}'
            html = html.replace("$$USER_NAME$$", contact.name if hasattr(contact, 'name') else contact.real_name)
            html = html.replace("$$PORTAL_LINK$$", chat_url)
            html = html.replace("$$HELP_TEXT$$", get_help_text(user))
            message_data = {
                "html": html,
                "subject": 'New message(s) on Leadguru',
                "recipient": user.email,
                "images": [ImageName.LOGO, ImageName.ARROW, ImageName.MAIL]
            }

        BackgroundJobRunner.submit(SendMailJob, SendMailJobData(**message_data))



import time
from abc import ABC
from random import randint
from typing import Optional, Any
from lgt_jobs.lgt_common.slack_client.web_client import SlackWebClient
from lgt_jobs.lgt_data.mongo_repository import DedicatedBotRepository
import logging as log
from pydantic import BaseModel
from lgt_jobs.basejobs import BaseBackgroundJobData, BaseBackgroundJob

"""
Send Slack Code
"""


class SendMassMessageSlackChannelJobData(BaseBackgroundJobData, BaseModel):
    text: str
    channel_ids: list[str]
    user_id: str
    files: Optional[list[Any]]
    source_id: str


class SendMassMessageSlackChannelJob(BaseBackgroundJob, ABC):
    @property
    def job_data_type(self) -> type:
        return SendMassMessageSlackChannelJobData

    def exec(self, data: SendMassMessageSlackChannelJobData):
        bot = DedicatedBotRepository().get_one(user_id=data.user_id, source_id=data.source_id, only_valid=True)
        if not bot:
            log.warning(f"[SendMassMessageSlackChannelJob]: "
                        f"Bot not found or has invalid creds, source id:{data.source_id}")
            return

        slack_client = SlackWebClient(bot.token, bot.cookies)
        attempts = 0

        for channel in data.channel_ids:
            while attempts < 5:
                attempts += 1
                try:
                    post_message_response = slack_client.post_message(to=channel, text=data.text)
                    if not post_message_response['ok']:
                        log.warning(f"[SendMassMessageSlackChannelJob]: "
                                    f"Failed to post message. Attempt {attempts}, bot id {bot.id},"
                                    f" channel id {channel}. Details {post_message_response}")
                    attempts = 0
                    break
                except:
                    log.warning(f"[SendMassMessageSlackChannelJob]: "
                                f"Failed attempt to send message. Attempt {attempts}, bot id {bot.id}")
                    time.sleep(randint(1, 3))
